# -*- coding: utf-8 -*-
"""
SCML (formerly SC3ML) events read and write support.

:author:
    EOST (École et Observatoire des Sciences de la Terre)
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import io
from pathlib import Path
import re

from lxml import etree

from obspy.core.util.decorator import deprecated
from obspy.io.quakeml.core import Pickler, Unpickler, _xml_doc_from_anything
from obspy.io.seiscomp.core import validate as validate_scml

# note that for schema 0.9 and below the following is NOT FIXED
# but for legacy reason remains as-is.
#
#  * 26.07.2024:
#    - Fix origin/confidenceEllipsoid conversion. The unit for
#      'semiMajorAxisLength', 'semiMinorAxisLength' and
#      'semiIntermediateAxisLength' is already meter and does not need a
#      conversion.

SCHEMA_VERSION = ['0.7', '0.8', '0.9', '0.10',
                  '0.11', '0.12', '0.13', '0.14']
# from version 0.14 onwards "sc3ml" is dropped
NEW_SCHEMA_VERSION = ['0.14']


# deprecated with 1.5.0, remove a few main releases later
@deprecated(
    'Deprecated and will be removed next release, use high level '
    'read_events(.., format="SCML") or if really needed _read_scml() instead')
def _read_sc3ml(*args, **kwargs):
    return _read_scml(*args, **kwargs)


def _read_scml(filename, id_prefix='smi:org.gfz.de/geofon/'):
    """
    Read a SeisComp XML file and returns a :class:`~obspy.core.event.Catalog`.

    An XSLT file is used to convert the SCML file to a QuakeML file. The
    catalog is then generated using the QuakeML module.

    .. warning::

        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.event.catalog.Catalog.write` method of an
        ObsPy :class:`~obspy.core.event.catalog.Catalog` object, call this
        instead.

    :type filename: str
    :param filename: SCML file to be read.
    :type id_prefix: str
    :param id_prefix: ID prefix. SCML does not enforce any particular ID
        restriction, this ID prefix allows to convert the IDs to a well
        formatted QuakeML ID. You can modify the default ID prefix with the
        reverse DNS name of your institute.
    :rtype: :class:`~obspy.core.event.Catalog`
    :return: An ObsPy Catalog object.

    .. rubric:: Example

    >>> from obspy import read_events
    >>> cat = read_events('/path/to/iris_events.sc3ml')
    >>> print(cat)
    2 Event(s) in Catalog:
    2011-03-11T05:46:24.120000Z | +38.297, +142.373
    2006-09-10T04:26:33.610000Z |  +9.614, +121.961
    """
    scml_doc = _xml_doc_from_anything(filename)

    match = re.match(
        r'{http://geofon\.gfz(?:-potsdam)?\.de/ns/seiscomp3?-schema/([-+]?'
        r'[0-9]*\.?[0-9]+)}', scml_doc.tag)

    try:
        version = match.group(1)
    except AttributeError:
        raise ValueError("Not a SCML compatible file or string.")
    else:
        if version not in SCHEMA_VERSION:
            message = ("Can't read SCML version %s, ObsPy can handle "
                       "versions [%s].") % (
                version, ', '.join(SCHEMA_VERSION))
            raise ValueError(message)

    xslt_filename = Path(__file__).parent / 'data'
    if version in NEW_SCHEMA_VERSION:
        xslt_filename = xslt_filename / f'scml_{version}__quakeml_1.2.xsl'
        # this doesn't need to be set necessarily, but may as well stay in tune
        id_prefix = 'smi:org.gfz.de/geofon/'
    else:
        xslt_filename = xslt_filename / f'sc3ml_{version}__quakeml_1.2.xsl'

    transform = etree.XSLT(etree.parse(str(xslt_filename)))
    quakeml_doc = transform(scml_doc,
                            ID_PREFIX=etree.XSLT.strparam(id_prefix))

    return Unpickler().load(io.BytesIO(quakeml_doc))


# deprecated with 1.5.0, remove a few main releases later
@deprecated(
    'Deprecated and will be removed next release, use high level '
    'Catalog.write(.., format="SCML") or if really needed _write_scml() '
    'instead')
def _write_sc3ml(*args, **kwargs):
    return _write_scml(*args, **kwargs)


def _write_scml(catalog, filename, validate=False, verbose=False,
                event_removal=False, version='0.12',
                **kwargs):  # @UnusedVariable
    """
    Write a SCML event file at desired version. Since a XSLT file is used to
    write the SCML file from a QuakeML file, the catalog is first
    converted in QuakeML.

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.event.catalog.Catalog.write` method of an
        ObsPy :class:`~obspy.core.event.catalog.Catalog` object, call this
        instead.

    :type catalog: :class:`~obspy.core.event.catalog.Catalog`
    :param catalog: The ObsPy Catalog object to write.
    :type filename: str or file
    :param filename: Filename to write or open file-like object
    :type validate: bool
    :param validate: If True, the final SCML file will be validated against
        the SCML schema file. Raises an AssertionError if the validation
        fails.
    :type verbose: bool
    :param verbose: Print validation error log if True.
    :type event_deletion: bool
    :param event_removal: If True, the event elements will be removed. This can
        be useful to associate origins with scevent when injecting SCML file
        into seiscomp.
    :type version: str
    :param version: SCML version to output (default is 0.12 ~ SC 5.0)
    """
    if version not in SCHEMA_VERSION:
        raise ValueError('%s is not a supported version. Use one of these '
                         'versions: [%s].'
                         % (version, ', '.join(SCHEMA_VERSION)))
    nsmap_ = getattr(catalog, "nsmap", {})
    quakeml_doc = Pickler(nsmap=nsmap_).dumps(catalog)
    xslt_filename = Path(__file__).parent / 'data'
    if version in NEW_SCHEMA_VERSION:
        xslt_filename = xslt_filename / f'quakeml_1.2__scml_{version}.xsl'
    else:
        xslt_filename = xslt_filename / f'quakeml_1.2__sc3ml_{version}.xsl'
    transform = etree.XSLT(etree.parse(str(xslt_filename)))
    scml_doc = transform(etree.parse(io.BytesIO(quakeml_doc)))

    # Remove events
    if event_removal:
        for event in scml_doc.xpath("//*[local-name()='event']"):
            event.getparent().remove(event)

    if validate and not validate_scml(io.BytesIO(scml_doc), verbose=verbose):
        raise AssertionError("The final SCML file did not pass validation.")

    # Open filehandler or use an existing file like object
    try:
        with open(filename, 'wb') as fh:
            fh.write(scml_doc)
    except TypeError:
        filename.write(scml_doc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test suite for the station handling.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2013
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""

import pytest

from obspy import read_inventory, UTCDateTime
from obspy.core.inventory import Station
from obspy.core.util.base import CatchAndAssertWarnings


@pytest.mark.usefixtures('ignore_numpy_errors')
class TestStation:
    """
    Tests the for :class:`~obspy.core.inventory.station.Station` class.
    """
    def test_station_response_plot(self, image_path):
        """
        Tests the response plot.
        """
        sta = read_inventory()[0][0]
        with CatchAndAssertWarnings():
            sta.plot(0.05, channel="*[NE]", outfile=image_path)

    def test_response_plot_degrees(self, image_path):
        """
        Tests the response plot.
        """
        sta = read_inventory()[0][0]
        with CatchAndAssertWarnings():
            sta.plot(0.05, channel="*[NE]", plot_degrees=True,
                     outfile=image_path)

    def test_len(self):
        """
        Tests the __len__ property.
        """
        sta = read_inventory()[0][0]

        assert len(sta) == len(sta.channels)
        assert len(sta) == 12

    def test_station_select(self):
        """
        Tests the select() method on station objects.
        """
        sta = read_inventory()[0][0]

        # Basic assertions to make sure the test data does not change.
        assert len(sta) == 12
        assert sta.code == "FUR"
        out = sorted(["%s.%s" % (_i.location_code, _i.code) for _i in sta])
        expected = (['.BHE', '.BHN', '.BHZ', '.HHE', '.HHN', '.HHZ',
                     '.LHE', '.LHN', '.LHZ', '.VHE', '.VHN', '.VHZ'])
        assert out == expected

        assert sta[0].code == "HHZ"
        # Manually set the end-date of the first one.
        sta[0].end_date = UTCDateTime(2010, 1, 1)

        # If nothing is given, nothing should change.
        sta_2 = sta.select()
        assert len(sta_2) == 12
        assert sta_2.code == "FUR"

        # Only select vertical channels.
        sta_2 = sta.select(channel="*Z")
        assert len(sta_2) == 4
        assert sta_2.code == "FUR"
        out = sorted(["%s.%s" % (_i.location_code, _i.code) for _i in sta_2])
        assert out == ['.BHZ', '.HHZ', '.LHZ', '.VHZ']

        # Only BH channels.
        sta_2 = sta.select(channel="BH?")
        assert len(sta_2) == 3
        assert sta_2.code == "FUR"

        out = sorted(["%s.%s" % (_i.location_code, _i.code) for _i in sta_2])
        assert out == ['.BHE', '.BHN', '.BHZ']

        # All location codes.
        sta_2 = sta.select(location="*")
        assert len(sta_2) == 12
        assert sta_2.code == "FUR"

        sta_2 = sta.select(location="")
        assert len(sta_2) == 12
        assert sta_2.code == "FUR"

        # None exist with this code.
        sta_2 = sta.select(location="10")
        assert len(sta_2) == 0
        assert sta_2.code == "FUR"

        # The time parameter selects channels active at that particular
        # time. All channels start 2006-12-16 and only the first ends in
        # 2010-1-1. All others don't have an end-date set.
        assert len(sta.select(time=UTCDateTime(2005, 1, 1))) == 0
        assert len(sta.select(time=UTCDateTime(2007, 1, 1))) == 12
        assert len(sta.select(time=UTCDateTime(2006, 12, 15))) == 0
        assert len(sta.select(time=UTCDateTime(2006, 12, 17))) == 12
        assert len(sta.select(time=UTCDateTime(2012, 1, 1))) == 11

        # Test starttime parameter.
        assert len(sta.select(starttime=UTCDateTime(2005, 1, 1))) == 12
        assert len(sta.select(starttime=UTCDateTime(2009, 1, 1))) == 12
        assert len(sta.select(starttime=UTCDateTime(2011, 1, 1))) == 11
        assert len(sta.select(starttime=UTCDateTime(2016, 1, 1))) == 11

        # Test endtime parameter.
        assert len(sta.select(endtime=UTCDateTime(2005, 1, 1))) == 0
        assert len(sta.select(endtime=UTCDateTime(2009, 1, 1))) == 12
        assert len(sta.select(endtime=UTCDateTime(2011, 1, 1))) == 12
        assert len(sta.select(endtime=UTCDateTime(2016, 1, 1))) == 12

        # Sampling rate parameter.
        assert len(sta.select(sampling_rate=33.0)) == 0
        assert len(sta.select(sampling_rate=100.0)) == 3
        assert len(sta.select(sampling_rate=20.0)) == 3
        assert len(sta.select(sampling_rate=1.0)) == 3
        assert len(sta.select(sampling_rate=0.1)) == 3

        out = sorted(["%s.%s" % (_i.location_code, _i.code) for _i
                     in sta.select(sampling_rate=100.0)])
        assert out == ['.HHE', '.HHN', '.HHZ']

        # Check tolerances.
        assert len(sta.select(sampling_rate=33.0 + 1E-6)) == 0
        assert len(sta.select(sampling_rate=100.0 + 1E-6)) == 3
        assert len(sta.select(sampling_rate=20.0 - 1E-6)) == 3
        assert len(sta.select(sampling_rate=1.0 + 1E-6)) == 3
        assert len(sta.select(sampling_rate=0.1 - 1E-6)) == 3

        # Artificially set different coordinates for a channel of RJOB.
        sta = read_inventory()[1][0]
        sta[0].latitude = 47.9
        sta[0].longitude = 12.9
        out = sta.select(
            minlatitude=47.8, maxlatitude=48,
            minlongitude=12.8, maxlongitude=13)
        assert len(out) == 1
        assert len(sta.select(
            latitude=47.95, longitude=12.95, maxradius=0.1)) == 1
        assert len(sta.select(
            latitude=47.95, longitude=12.95, minradius=0.1)) == 2
        assert len(sta.select(
            latitude=47.95, longitude=12.95,
            minradius=0.08, maxradius=0.1)) == 0

    def test_warn_identifier_invalid_uri_syntax(self):
        """
        Tests the warning on Identifiers getting set with an invalid URI (not
        having scheme-colon-path)
        """
        sta = Station(code='A', latitude=1, longitude=1, elevation=1)
        invalid_uri = "this-has-no-URI-scheme-and-no-colon"
        msg = f"Given string seems to not be a valid URI: '{invalid_uri}'"
        with CatchAndAssertWarnings(expected=[(UserWarning, msg)]):
            sta.identifiers = [invalid_uri]



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------
# Filename: source.py
#  Purpose: Computes and plots radiation patterns
# ---------------------------------------------------------------------
"""
Functions to compute and plot radiation patterns

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""

import numpy as np

from itertools import chain

from obspy.core.event.source import farfield
from obspy.imaging.scripts.mopad import MomentTensor, BeachBall
from obspy.imaging.mopad_wrapper import beach
from obspy.core.util import CARTOPY_VERSION
if CARTOPY_VERSION:
    HAS_CARTOPY = True
else:
    HAS_CARTOPY = False


def _setup_figure_and_axes(kind, fig=None, subplot_size=4.0, **kwargs):
    """
    Setup figure for Event plot.

    :param kind: A list of strings or nested list of strings, see
        :meth:`obspy.core.event.event.Event.plot`.
    :type kind: list[str] or list[list[str]]
    :type subplot_size: float
    :param subplot_size: Width/height of one single subplot cell in inches.
    :rtype: tuple
    :returns: A 3-tuple with a :class:`~matplotlib.figure.Figure`, a list of
        :class:`~matplotlib.axes.Axes` and a list of strings with corresponding
        plotting options for each axes (see
        :meth:`obspy.core.event.event.Event.plot`, parameter `kind`).
    """
    import matplotlib.pyplot as plt
    # restrict potential fails on matplotlib 1.5.1 (e.g. Ubuntu xenial) owed to
    # matplotlib/matplotlib#6537 to routines that actually use Axes3D, for that
    # reason do the Axes3D import inside this routine.
    from mpl_toolkits.mplot3d import Axes3D  # NOQA
    # make 2d layout of kind parameter
    if isinstance(kind[0], (list, tuple)):
        nrows = len(kind)
        ncols = max([len(k) for k in kind])
    else:
        nrows, ncols = 1, len(kind)
        kind = [kind]
    figsize = (ncols * subplot_size, nrows * subplot_size)
    if not fig:
        fig = plt.figure(figsize=figsize, facecolor='white')
    # get tuple of "kind" options and corresponding axes
    kind_ = []
    axes = []
    for i, row in enumerate(kind):
        ncols_ = len(row)
        for j, kind__ in enumerate(row):
            kind_.append(kind__)
            kwargs["adjustable"] = "datalim"
            if kind__ in ("p_quiver", "p_sphere", "s_quiver", "s_sphere"):
                kwargs["projection"] = "3d"
                kwargs["aspect"] = "auto"
            if kind__ in ("ortho", "local", "global"):
                import cartopy.crs as ccrs
                lats = []
                lons = []
                if "events" in kwargs:
                    _cat = kwargs.pop("events")
                    for event in _cat:
                        origin = event.preferred_origin() or event.origins[0]
                        lats.append(origin.latitude)
                        lons.append(origin.longitude)
                    lat_0 = round(np.mean(lats), 4)
                    lon_0 = round(np.mean(lons), 4)
                else:
                    lat_0 = 0.0
                    lon_0 = 0.0
                if kind__ == "ortho":
                    kwargs["projection"] = ccrs.Orthographic(
                        central_longitude=lon_0,
                        central_latitude=lat_0)
                elif kind__ == "global":
                    kwargs["projection"] = ccrs.Mollweide(
                        central_longitude=lon_0
                    )
                else:
                    kwargs["projection"] = ccrs.AlbersEqualArea(
                        central_longitude=lon_0,
                        central_latitude=lat_0
                    )
                kwargs["aspect"] = "equal"
            else:  # equal aspect never worked on 3d plot, see mpl #13474
                kwargs["aspect"] = "auto"
            ax = fig.add_subplot(nrows, ncols_, i * ncols_ + j + 1, **kwargs)
            axes.append(ax)
    return fig, axes, kind_


def plot_radiation_pattern(
        mt, kind=['p_sphere', 'beachball'], coordinate_system='RTP',
        p_sphere_direction='inwards', fig=None, show=True):
    """
    Plot the P/S farfield radiation pattern on a unit sphere grid.

    The calculations are based on [Aki1980]_ eq. 4.29.

    :param mt: Focal mechanism NM x 6 (M11, M22, M33, M12, M13, M23 - the
        six independent components of the moment tensor, where the coordinate
        system is 1,2,3 = Up,South,East which equals r,theta,phi -
        Harvard/Global CMT convention). The relation to [Aki1980]_
        x,y,z equals North,East,Down convention is as follows: Mrr=Mzz,
        Mtt=Mxx, Mpp=Myy, Mrt=Mxz, Mrp=-Myz, Mtp=-Mxy.
    :param kind: One of:

        * **(A)** A list of strings or nested list of strings for a matplotlib
          plot (for details see :meth:`obspy.core.event.event.Event.plot`).
        * **(B)** ``"mayavi"``: uses the mayavi library.
        * **(C)** ``"vtk"``: This vtk option writes two vtk files to the
          current working directory. ``rpattern.vtk`` contains the p and s
          wave farfield vector field. ``beachlines.vtk`` contains the nodal
          lines of the radiation pattern. A vtk glyph filter should be applied
          to the vector field (e.g. in ParaView) to visualize it.

    :type fig: :class:`matplotlib.figure.Figure`
    :param fig: Figure instance to use.
    :type show: bool
    :param show: Whether to show the figure after plotting or not. Can be
        used to do further customization of the plot before showing it.
    :returns: Matplotlib figure or ``None`` (if ``kind`` is ``"mayavi"`` or
        ``"vtk"``)
    """
    import matplotlib.pyplot as plt

    # reoorder all moment tensors to NED and RTP convention
    # name : COMPONENT              : NED sign and index
    # NED  : NN, EE, DD, NE, ND, ED : [0, 1, 2, 3, 4, 5]
    # USE  : UU, SS, EE, US, UE, SE : [1, 2, 0, -5, 3, -4]
    # RTP  : RR, TT, PP, RT, RP, TP : [1, 2, 0, -5, 3, -4]
    # DSE  : DD, SS, EE, DS, DE, SE : [1, 2, 0, -5, -3, 4]
    if coordinate_system == 'RTP' or coordinate_system == 'USE':
        signs = [1, 1, 1, -1, 1, -1]
        indices = [1, 2, 0, 5, 3, 4]
        ned_mt = [sign * mt[ind] for sign, ind in zip(signs, indices)]
        rtp_mt = mt
    # the moment tensor has to be converted to
    # RTP/USE coordinates as well because the beachball routine relies
    # on it.
    # elif coordinate_system == 'DSE':
    #     signs = [1, 1, 1, -1, -1, 1]
    #     indices = [1, 2, 0, 5, 3, 4]
    #     ned_mt = [sign * mt[ind] for sign, ind in zip(signs, indices)]
    # elif coordinate_system == 'NED':
    #     ned_mt = mt
    else:
        msg = 'moment tensor in {:s} coordinates not implemented yet'
        raise NotImplementedError(msg.format(coordinate_system))

    # matplotlib plotting is triggered when kind is a list of strings
    if isinstance(kind, (list, tuple)):
        if not fig:
            fig, axes, kind = _setup_figure_and_axes(kind, fig=fig)
        else:
            axes = fig.axes
            if len(kind) == 1:
                kind = kind
            else:
                kind = list(chain(*kind))
        for ax, kind_ in zip(axes, kind):
            if kind_ is None:
                continue
            elif kind_ == 'p_quiver':
                _plot_radiation_pattern_quiver(ax, ned_mt, type="P")
            elif kind_ == 'p_sphere':
                _plot_radiation_pattern_sphere(
                    ax, ned_mt, type="P",
                    p_sphere_direction=p_sphere_direction)
            elif kind_ == 's_quiver':
                _plot_radiation_pattern_quiver(ax, ned_mt, type="S")
            elif kind_ == 's_sphere':
                _plot_radiation_pattern_sphere(ax, ned_mt, type="S")
            elif kind_ == 'beachball':
                ax.spines['left'].set_position('center')
                ax.spines['right'].set_color('none')
                ax.spines['bottom'].set_position('center')
                ax.spines['top'].set_color('none')
                _plot_beachball(ax, rtp_mt)
        # see https://github.com/SciTools/cartopy/issues/1207
        fig.canvas.draw()
        fig.tight_layout(pad=0.1)
        if show:
            plt.show()
        return fig

    elif kind == 'mayavi':
        _plot_radiation_pattern_mayavi(ned_mt)

    elif kind == 'vtk':
        # this saves two files, one with the vector field and one
        # with the nodal lines of the beachball
        fname_rpattern = 'rpattern.vtk'
        fname_beachlines = 'beachlines.vtk'
        _write_radiation_pattern_vtk(
            ned_mt, fname_rpattern=fname_rpattern,
            fname_beachlines=fname_beachlines)

    else:
        raise NotImplementedError('{:s} not implemented yet'.format(kind))


def _plot_radiation_pattern_sphere(
        ax3d, ned_mt, type, p_sphere_direction='inwards'):
    """
    Private function that plots a radiation pattern sphere into an
    :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D`.

    :type ax3d: :class:`mpl_toolkits.mplot3d.axes3d.Axes3D`
    :param ax3d: matplotlib Axes3D object
    :param ned_mt: moment tensor in NED convention
    :param p_sphere_direction: If this is 'inwards', the tension regions of the
        beachball deform the radiation sphere inwards. If 'outwards' it deforms
        outwards.
    :param type: 'P' or 'S' (P or S wave).
    """
    import matplotlib.pyplot as plt
    type = type.upper()
    if type not in ("P", "S"):
        msg = ("type must be 'P' or 'S'")
        raise ValueError(msg)
    is_p_wave = type == "P"

    # generate spherical mesh that is aligned with the moment tensor null
    # axis. MOPAD should use NED coordinate system to avoid internal
    # coordinate transformations
    mtensor = MomentTensor(ned_mt, system='NED')

    # use the most isolated eigenvector as axis of symmetry
    evecs = mtensor.get_eigvecs()
    evals = np.abs(mtensor.get_eigvals())**2
    evals_dev = np.abs(evals - np.mean(evals))
    if is_p_wave:
        if p_sphere_direction == 'outwards':
            evec_max = evecs[np.argmax(evals_dev)]
        elif p_sphere_direction == 'inwards':
            evec_max = evecs[np.argmax(evals)]
    else:
        evec_max = evecs[np.argmax(evals_dev)]
    orientation = np.ravel(evec_max)

    # get a uv sphere that is oriented along the moment tensor axes
    ntheta, nphi = 100, 100
    points = _oriented_uv_sphere(ntheta=ntheta, nphi=nphi,
                                 orientation=orientation)
    sshape = (ntheta, nphi)

    # get radiation pattern
    if is_p_wave:
        disp = farfield(ned_mt, points, type="P")
        magn = np.sum(disp * points, axis=0)
        cmap = plt.get_cmap('bwr')
        norm = plt.Normalize(-1, 1)
    else:
        disp = farfield(ned_mt, points, type="S")
        magn = np.sqrt(np.sum(disp * disp, axis=0))
        cmap = plt.get_cmap('Greens')
        norm = plt.Normalize(0, 1)
    magn /= np.max(np.abs(magn))

    # compute colours and displace points along normal
    if is_p_wave:
        if p_sphere_direction == 'outwards':
            points *= (1. + np.abs(magn) / 2.)
        elif p_sphere_direction == 'inwards':
            points *= (1. + magn / 2.)
    else:
        points *= (1. + magn / 2.)
    colors = np.array([cmap(norm(val)) for val in magn])
    colors = colors.reshape(ntheta, nphi, 4)

    x = points[0].reshape(sshape)
    y = points[1].reshape(sshape)
    z = points[2].reshape(sshape)

    # plot 3d radiation pattern
    ax3d.plot_surface(x, y, z, rstride=4, cstride=4, facecolors=colors)
    ax3d.set(xlim=(-1.5, 1.5), ylim=(-1.5, 1.5), zlim=(-1.5, 1.5),
             xticks=[-1, 1], yticks=[-1, 1], zticks=[-1, 1],
             xticklabels=['South', 'North'],
             yticklabels=['West', 'East'],
             zticklabels=['Up', 'Down'],
             title='{} wave farfield'.format(type))
    ax3d.view_init(elev=-110., azim=0.)


def _plot_radiation_pattern_quiver(ax3d, ned_mt, type):
    """
    Private routine that plots the wave farfield into an
    :class:`~mpl_toolkits.mplot3d.axes3d.Axes3D` object

    :type ax3d: :class:`mpl_toolkits.mplot3d.axes3d.Axes3D`
    :param ax3d: matplotlib Axes3D object
    :param ned_mt: the 6 comp moment tensor in NED orientation
    :type type: str
    :param type: 'P' or 'S' (P or S wave).
    """
    import matplotlib.pyplot as plt

    type = type.upper()
    if type not in ("P", "S"):
        msg = ("type must be 'P' or 'S'")
        raise ValueError(msg)
    is_p_wave = type == "P"

    # precompute even spherical grid and directional cosine array
    points = _equalarea_spherical_grid(nlat=14)

    if is_p_wave:
        # get radiation pattern
        disp = farfield(ned_mt, points, type="P")
        # normalized magnitude:
        magn = np.sum(disp * points, axis=0)
        magn /= np.max(np.abs(magn))
        cmap = plt.get_cmap('bwr')
    else:
        # get radiation pattern
        disp = farfield(ned_mt, points, type="S")
        # normalized magnitude (positive only):
        magn = np.sqrt(np.sum(disp * disp, axis=0))
        magn /= np.max(np.abs(magn))
        cmap = plt.get_cmap('Greens')

    # plot
    # there is a mlab3d bug that quiver vector colors and lengths
    # can only be changed if we plot each arrow independently
    for loc, vec, mag in zip(points.T, disp.T, magn.T):
        norm = plt.Normalize(-1., 1.)
        color = cmap(norm(mag))
        if is_p_wave:
            loc *= (1. + mag / 2.)
            length = abs(mag) / 2.0
        else:
            length = abs(mag) / 5.0
        ax3d.quiver(loc[0], loc[1], loc[2], vec[0], vec[1], vec[2],
                    length=length, color=color)
    ax3d.set(xlim=(-1.5, 1.5), ylim=(-1.5, 1.5), zlim=(-1.5, 1.5),
             xticks=[-1, 1], yticks=[-1, 1], zticks=[-1, 1],
             xticklabels=['South', 'North'],
             yticklabels=['West', 'East'],
             zticklabels=['Up', 'Down'],
             title='{} wave farfield'.format(type))
    ax3d.view_init(elev=-110., azim=0.)


def _plot_beachball(ax2d, rtp_mt):
    """
    Private function that plots a beachball into a 2d matplotlib
    :class:`~matplotlib.axes.Axes`.

    :type ax2d: :class:`matplotlib.axes.Axes`
    :param ax2d: 2d matplotlib Axes
    :param ax2d: matplotlib Axes3D object
    :param rtp_mt: moment tensor in RTP convention
    """
    import matplotlib.pyplot as plt
    norm = plt.Normalize(-1., 1.)
    cmap = plt.get_cmap('bwr')
    bball = beach(rtp_mt, xy=(0, 0), width=50, facecolor=cmap(norm(0.7)),
                  bgcolor=cmap(norm(-0.7)))

    ax2d.add_collection(bball)
    ax2d.set(xlim=(-50, 50), ylim=(-50, 50),
             xticks=(-40, 40), yticks=(-40, 40),
             xticklabels=('West', 'East'),
             yticklabels=('South', 'North'),
             title='lower hemisphere stereographical projection')


def _plot_radiation_pattern_mayavi(ned_mt):
    """
    Plot the radiation pattern using MayaVi.

    This private function uses the mayavi (vtk) library to plot the radiation
    pattern to screen. Note that you might have to set the QT_API environmental
    variable to e.g. export QT_API=pyqt that mayavi works properly.

    :param ned_mt: moment tensor in NED convention
    """
    # use mayavi if possible.
    try:
        from mayavi import mlab
    except Exception as err:
        print(err)
        msg = ("ObsPy failed to import MayaVi. "
               "You need to install the mayavi module "
               "(e.g. 'conda install mayavi', 'pip install mayavi'). "
               "If it is installed and still doesn't work, "
               "try setting the environmental variable QT_API to "
               "pyqt (e.g. export QT_API=pyqt) before running the "
               "code. Another option is to avoid mayavi and "
               "directly use kind='vtk' for vtk file output of the "
               "radiation pattern that can be used by external "
               "software like ParaView")
        raise ImportError(msg)

    # get mopad moment tensor
    mopad_mt = MomentTensor(ned_mt, system='NED')
    bb = BeachBall(mopad_mt, npoints=200)
    bb._setup_BB(unit_circle=False)

    # extract the coordinates of the nodal lines
    neg_nodalline = bb._nodalline_negative
    pos_nodalline = bb._nodalline_positive

    # add the first point to the end to close the nodal line
    neg_nodalline = np.hstack((neg_nodalline, neg_nodalline[:, 0][:, None]))
    pos_nodalline = np.hstack((pos_nodalline, pos_nodalline[:, 0][:, None]))

    # plot radiation pattern and nodal lines
    points = _equalarea_spherical_grid(nlat=20)
    dispp = farfield(ned_mt, points, type="P")
    disps = farfield(ned_mt, points, type="S")

    # get vector lengths
    normp = np.sum(dispp * points, axis=0)
    normp /= np.max(np.abs(normp))

    norms = np.sqrt(np.sum(disps * disps, axis=0))
    norms /= np.max(np.abs(norms))

    # make sphere to block view to the other side of the beachball
    rad = 0.8
    pi = np.pi
    cos = np.cos
    sin = np.sin
    phi, theta = np.mgrid[0:pi:101j, 0:2 * pi:101j]

    x = rad * sin(phi) * cos(theta)
    y = rad * sin(phi) * sin(theta)
    z = rad * cos(phi)

    # p wave radiation pattern
    mlab.figure(size=(800, 800), bgcolor=(0, 0, 0))
    pts1 = mlab.quiver3d(points[0], points[1], points[2],
                         dispp[0], dispp[1], dispp[2],
                         scalars=normp, vmin=-1., vmax=1.)
    pts1.glyph.color_mode = 'color_by_scalar'
    mlab.plot3d(*neg_nodalline, color=(0, 0.5, 0), tube_radius=0.01)
    mlab.plot3d(*pos_nodalline, color=(0, 0.5, 0), tube_radius=0.01)
    mlab.mesh(x, y, z, color=(0, 0, 0))

    # s wave radiation pattern
    mlab.figure(size=(800, 800), bgcolor=(0, 0, 0))
    pts2 = mlab.quiver3d(points[0], points[1], points[2],
                         disps[0], disps[1], disps[2], scalars=norms,
                         vmin=-0., vmax=1.)
    pts2.glyph.color_mode = 'color_by_scalar'
    mlab.plot3d(*neg_nodalline, color=(0, 0.5, 0), tube_radius=0.01)
    mlab.plot3d(*pos_nodalline, color=(0, 0.5, 0), tube_radius=0.01)
    mlab.mesh(x, y, z, color=(0, 0, 0))

    mlab.show()


def _write_radiation_pattern_vtk(
        ned_mt, fname_rpattern='rpattern.vtk',
        fname_beachlines='beachlines.vtk'):
    # output a vtkfile that can for exampled be displayed by ParaView
    mtensor = MomentTensor(ned_mt, system='NED')
    bb = BeachBall(mtensor, npoints=200)
    bb._setup_BB(unit_circle=False)

    # extract the coordinates of the nodal lines
    neg_nodalline = bb._nodalline_negative
    pos_nodalline = bb._nodalline_positive

    # plot radiation pattern and nodal lines
    points = _equalarea_spherical_grid()
    ndim, npoints = points.shape
    dispp = farfield(ned_mt, points, type="P")
    disps = farfield(ned_mt, points, type="S")

    # write vector field
    with open(fname_rpattern, 'w') as vtk_file:
        vtk_header = '# vtk DataFile Version 2.0\n' + \
                     'radiation pattern vector field\n' + \
                     'ASCII\n' + \
                     'DATASET UNSTRUCTURED_GRID\n' + \
                     'POINTS {:d} float\n'.format(npoints)

        vtk_file.write(vtk_header)
        # write point locations
        for x, y, z in np.transpose(points):
            vtk_file.write('{:.3e} {:.3e} {:.3e}\n'.format(x, y, z))
        # write vector field
        vtk_file.write('POINT_DATA {:d}\n'.format(npoints))
        vtk_file.write('VECTORS s_radiation float\n')
        for x, y, z in np.transpose(disps):
            vtk_file.write('{:.3e} {:.3e} {:.3e}\n'.format(x, y, z))
        vtk_file.write('VECTORS p_radiation float\n')
        for x, y, z in np.transpose(dispp):
            vtk_file.write('{:.3e} {:.3e} {:.3e}\n'.format(x, y, z))

    # write nodal lines
    with open(fname_beachlines, 'w') as vtk_file:
        npts_neg = neg_nodalline.shape[1]
        npts_pos = pos_nodalline.shape[1]
        npts_tot = npts_neg + npts_pos
        vtk_header = '# vtk DataFile Version 2.0\n' + \
                     'beachball nodal lines\n' + \
                     'ASCII\n' + \
                     'DATASET UNSTRUCTURED_GRID\n' + \
                     'POINTS {:d} float\n'.format(npts_tot)

        vtk_file.write(vtk_header)
        # write point locations
        for x, y, z in np.transpose(neg_nodalline):
            vtk_file.write('{:.3e} {:.3e} {:.3e}\n'.format(x, y, z))
        for x, y, z in np.transpose(pos_nodalline):
            vtk_file.write('{:.3e} {:.3e} {:.3e}\n'.format(x, y, z))

        # write line segments
        vtk_file.write('\nCELLS 2 {:d}\n'.format(npts_tot + 4))

        ipoints = list(range(0, npts_neg)) + [0]
        vtk_file.write('{:d} '.format(npts_neg + 1))
        for ipoint in ipoints:
            if ipoint % 30 == 29:
                vtk_file.write('\n')
            vtk_file.write('{:d} '.format(ipoint))
        vtk_file.write('\n')

        ipoints = list(range(0, npts_pos)) + [0]
        vtk_file.write('{:d} '.format(npts_pos + 1))
        for ipoint in ipoints:
            if ipoint % 30 == 29:
                vtk_file.write('\n')
            vtk_file.write('{:d} '.format(ipoint + npts_neg))
        vtk_file.write('\n')

        # cell types. 4 means cell type is a poly_line
        vtk_file.write('\nCELL_TYPES 2\n')
        vtk_file.write('4\n4')


# ===== SUPPORT FUNCTIONS FOR SPHERICAL MESHES ETC STARTING HERE:
def _oriented_uv_sphere(ntheta=100, nphi=100, orientation=[0., 0., 1.]):
    """
    Returns a uv sphere (equidistant lat/lon grid) with its north-pole rotated
    to the input axis. It returns the spherical grid points that can be used to
    generate a QuadMesh on the sphere for surface plotting.

    :param nlat: number of latitudinal grid points (default = 100)
    :param nphi: number of longitudinal grid points (default = 100)
    :param orientation: axis of the north-pole of the sphere
                        (default = [0, 0, 1])
    """
    # make rotation matrix (after numpy mailing list)
    zaxis = np.array([0., 0., 1.])
    raxis = np.cross(orientation, zaxis)  # rotate z axis to null
    raxis_norm = np.linalg.norm(raxis)
    if raxis_norm < 1e-10:  # check for zero or 180 degree rotation
        rotmtx = np.eye(3, dtype=np.float64)
    else:
        raxis /= raxis_norm

        # angle between z and null
        angle = np.arccos(np.dot(zaxis, orientation))

        eye = np.eye(3, dtype=np.float64)
        raxis2 = np.outer(raxis, raxis)
        skew = np.array([[0, raxis[2], -raxis[1]],
                         [-raxis[2], 0, raxis[0]],
                         [raxis[1], -raxis[0], 0]])

        rotmtx = (raxis2 + np.cos(angle) * (eye - raxis2) +
                  np.sin(angle) * skew)

    # make uv sphere that is aligned with z-axis
    ntheta, nphi = 100, 100
    u = np.linspace(0, 2 * np.pi, nphi)
    v = np.linspace(0, np.pi, ntheta)

    x = np.outer(np.cos(u), np.sin(v))
    y = np.outer(np.sin(u), np.sin(v))
    z = np.outer(np.ones(np.size(u)), np.cos(v))

    # ravel point array and rotate them to the null axis
    points = np.vstack((x.flatten(), y.flatten(), z.flatten()))
    points = np.dot(rotmtx, points)
    return points


def _equalarea_spherical_grid(nlat=30):
    """
    Generates a simple spherical equalarea grid that adjust the number of
    longitude samples to the latitude. This grid is useful to plot vectors on
    the sphere but not surfaces.

    :param nlat: number of nodes in lat direction. The number of
                 nodes in lon direction is 2*nlat+1 at the equator
    """

    ndim = 3
    colats = np.linspace(0., np.pi, nlat)
    norms = np.sin(colats)
    # Scale number of point with latitude.
    nlons = (2 * nlat * norms + 1).astype(np.int_)

    # make colat/lon grid
    colatgrid, longrid = [], []
    for ilat in range(nlat):
        nlon = nlons[ilat]
        dlon = 2. * np.pi / nlon
        lons = np.linspace(0. + dlon / 2., 2. * np.pi - dlon / 2., nlon)
        for ilon in range(nlon):
            colatgrid.append(colats[ilat])
            longrid.append(lons[ilon])
    npoints = len(longrid)

    # get cartesian coordinates of spherical grid
    points = np.empty((ndim, npoints))
    points[0] = np.sin(colatgrid) * np.cos(longrid)
    points[1] = np.sin(colatgrid) * np.sin(longrid)
    points[2] = np.cos(colatgrid)

    return points


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
The polarization.core test suite.
"""
import numpy as np
import pytest
from scipy import signal

import obspy
from obspy.signal import polarization, util


def _create_test_data():
    """
    Test data used for some polarization tests.
    :return:
    """
    x = np.arange(0, 2048 / 20.0, 1.0 / 20.0)
    x *= 2. * np.pi
    y = np.cos(x)
    tr_z = obspy.Trace(data=y)
    tr_z.stats.sampling_rate = 20.
    tr_z.stats.starttime = obspy.UTCDateTime('2014-03-01T00:00')
    tr_z.stats.station = 'POLT'
    tr_z.stats.channel = 'HHZ'
    tr_z.stats.network = 'XX'

    tr_n = tr_z.copy()
    tr_n.data *= 2.
    tr_n.stats.channel = 'HHN'
    tr_e = tr_z.copy()
    tr_e.stats.channel = 'HHE'

    sz = obspy.Stream()
    sz.append(tr_z)
    sz.append(tr_n)
    sz.append(tr_e)
    sz.sort(reverse=True)

    return sz


class TestPolarization():
    """
    Test cases for polarization analysis
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup_data(self, testdata):
        # setting up sliding window data
        data_z = np.loadtxt(testdata['MBGA_Z.ASC'])
        data_e = np.loadtxt(testdata['MBGA_E.ASC'])
        data_n = np.loadtxt(testdata['MBGA_N.ASC'])
        n = 256
        fs = 75
        inc = int(0.05 * fs)
        self.data_win_z, self.nwin, self.no_win = \
            util.enframe(data_z, signal.windows.hamming(n), inc)
        self.data_win_e, self.nwin, self.no_win = \
            util.enframe(data_e, signal.windows.hamming(n), inc)
        self.data_win_n, self.nwin, self.no_win = \
            util.enframe(data_n, signal.windows.hamming(n), inc)
        # global test input
        self.fk = [2, 1, 0, -1, -2]
        self.norm = pow(np.max(data_z), 2)
        self.res = np.loadtxt(testdata['3cssan.hy.1.MBGA_Z'])

    def test_polarization(self):
        """
        windowed data
        """
        pol = polarization.eigval(self.data_win_e, self.data_win_n,
                                  self.data_win_z, self.fk, self.norm)
        rms = np.sqrt(np.sum((pol[0] - self.res[:, 34]) ** 2) /
                      np.sum(self.res[:, 34] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[1] - self.res[:, 35]) ** 2) /
                      np.sum(self.res[:, 35] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[2] - self.res[:, 36]) ** 2) /
                      np.sum(self.res[:, 36] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[3] - self.res[:, 40]) ** 2) /
                      np.sum(self.res[:, 40] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[4] - self.res[:, 42]) ** 2) /
                      np.sum(self.res[:, 42] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[5][:, 0] - self.res[:, 37]) ** 2) /
                      np.sum(self.res[:, 37] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[5][:, 1] - self.res[:, 38]) ** 2) /
                      np.sum(self.res[:, 38] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[5][:, 2] - self.res[:, 39]) ** 2) /
                      np.sum(self.res[:, 39] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[6] - self.res[:, 41]) ** 2) /
                      np.sum(self.res[:, 41] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((pol[7] - self.res[:, 43]) ** 2) /
                      np.sum(self.res[:, 43] ** 2))
        assert rms < 1.0e-5

    def test_polarization_1d(self):
        """
        1 dimenstional input --- regression test case for bug #919
        """
        pol = polarization.eigval(self.data_win_e[100, :],
                                  self.data_win_n[100, :],
                                  self.data_win_z[100, :],
                                  self.fk, self.norm)
        pol_5_ref = [2.81387533e-04, 3.18409580e-04, 6.74030846e-04,
                     5.55067015e-01, 4.32938188e-01]
        assert np.allclose(np.concatenate(pol[:5]), pol_5_ref)

    def test_polarization_pm(self):
        st = _create_test_data()
        t = st[0].stats.starttime
        e = st[0].stats.endtime
        wlen = 10.0
        wfrac = 0.1

        out = polarization.polarization_analysis(
            st, win_len=wlen, win_frac=wfrac, frqlow=1.0, frqhigh=5.0,
            verbose=False, stime=t, etime=e, method="pm",
            var_noise=0.0)

        # all values should be equal for the test data, so check first value
        # and make sure all values are almost equal
        assert out["timestamp"][0] == 1393632005.0
        assert out["timestamp"][0] == t + wlen / 2.0
        assert round(abs(out["azimuth"][0]-26.56505117707799), 7) == 0
        assert round(abs(out["incidence"][0]-65.905157447889309), 7) == 0
        assert round(abs(out["azimuth_error"][0]-0.000000), 7) == 0
        assert round(abs(out["incidence_error"][0]-0.000000), 7) == 0
        for key in ["azimuth", "incidence"]:
            got = out[key]
            assert np.allclose(got / got[0], np.ones_like(got), rtol=1e-4)
        for key in ["azimuth_error", "incidence_error"]:
            got = out[key]
            expected = np.empty_like(got)
            expected.fill(got[0])
            assert np.allclose(got, expected, rtol=1e-4, atol=1e-16)
        assert np.allclose(out["timestamp"] - out["timestamp"][0],
                           np.arange(0, 92, 1))

    def test_polarization_flinn(self):
        st = _create_test_data()
        t = st[0].stats.starttime
        e = st[0].stats.endtime
        wlen = 10.0
        wfrac = 0.1

        out = polarization.polarization_analysis(
            st, win_len=wlen, win_frac=wfrac, frqlow=1.0, frqhigh=5.0,
            verbose=False, stime=t, etime=e,
            method="flinn", var_noise=0.0)

        # all values should be equal for the test data, so check first value
        # and make sure all values are almost equal
        assert out["timestamp"][0] == 1393632005.0
        assert out["timestamp"][0] == t + wlen / 2.0
        assert round(abs(out["azimuth"][0]-26.56505117707799), 7) == 0
        assert round(abs(out["incidence"][0]-65.905157447889309), 7) == 0
        assert round(abs(out["rectilinearity"][0]-1.000000), 7) == 0
        assert round(abs(out["planarity"][0]-1.000000), 7) == 0
        for key in ["azimuth", "incidence", "rectilinearity", "planarity"]:
            got = out[key]
            assert np.allclose(got / got[0], np.ones_like(got), rtol=1e-4)
        assert np.allclose(out["timestamp"] - out["timestamp"][0],
                           np.arange(0, 92, 1))

    def test_polarization_vidale(self):
        st = _create_test_data()
        t = st[0].stats.starttime
        e = st[0].stats.endtime

        out = polarization.polarization_analysis(
            st, win_len=10.0, win_frac=0.1, frqlow=1.0, frqhigh=5.0,
            verbose=False, stime=t, etime=e,
            method="vidale", var_noise=0.0)

        # all values should be equal for the test data, so check first value
        # and make sure all values are almost equal
        assert out["timestamp"][0] == 1393632003.0
        assert round(abs(out["azimuth"][0]-26.56505117707799), 7) == 0
        assert round(abs(out["incidence"][0]-65.905157447889309), 7) == 0
        assert round(abs(out["rectilinearity"][0]-1.000000), 7) == 0
        assert round(abs(out["planarity"][0]-1.000000), 7) == 0
        assert round(abs(out["ellipticity"][0]-3.8195545129768958e-06), 7) == 0
        for key in ["azimuth", "incidence", "rectilinearity", "planarity",
                    "ellipticity"]:
            got = out[key]
            assert np.allclose(got / got[0], np.ones_like(got), rtol=1e-4)
        assert np.allclose(out["timestamp"] - out["timestamp"][0],
                           np.arange(0, 97.85, 0.05), rtol=1e-5)



# -*- coding: utf-8 -*-
import numpy as np

from obspy import Trace, UTCDateTime, read
from obspy.io.ascii.core import (_determine_dtype, _is_slist, _is_tspair,
                                 _read_slist, _read_tspair, _write_slist,
                                 _write_tspair)
from obspy.core.util import NamedTemporaryFile
import pytest


class TestASCII():
    """
    """
    def test_is_slist_file(self, testdata, datapath):
        """
        Testing SLIST file format.
        """
        testfile = testdata['slist.ascii']
        assert _is_slist(testfile)
        testfile = testdata['slist_2_traces.ascii']
        assert _is_slist(testfile)
        testfile = testdata['tspair.ascii']
        assert not _is_slist(testfile)
        # not existing file should fail
        testfile = datapath / 'xyz'
        assert not _is_slist(testfile)

    def test_read_slist_file_single_trace(self, testdata):
        """
        Read SLIST file test via obspy.core.ascii._read_slist.
        """
        testfile = testdata['slist.ascii']
        # read
        stream = _read_slist(testfile)
        stream.verify()
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 635
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[0].data[0:4], data)
        # check last 4 samples
        data = [761, 755, 748, 746]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data)

    def test_read_slist_file_multiple_traces(self, testdata):
        """
        Read SLIST file test via obspy.core.ascii._read_slist.
        """
        testfile = testdata['slist_2_traces.ascii']
        # read
        stream = _read_slist(testfile)
        stream.verify()
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 635
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[0].data[0:4], data)
        # check last 4 samples
        data = [761, 755, 748, 746]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data)
        # second trace
        assert stream[1].stats.network == 'XX'
        assert stream[1].stats.station == 'TEST'
        assert stream[1].stats.location == ''
        assert stream[1].stats.channel == 'BHE'
        assert stream[1].stats.sampling_rate == 40.0
        assert stream[1].stats.npts == 630
        assert stream[1].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[1].stats.calib == 1.0e-00
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[1].data[0:4], data)
        # check last 4 samples
        data = [781, 785, 778, 772]
        np.testing.assert_array_almost_equal(stream[1].data[-4:], data)

    def test_read_slist_file_head_only(self, testdata):
        """
        Read SLIST file test via obspy.core.ascii._read_slist.
        """
        testfile = testdata['slist.ascii']
        # read
        stream = _read_slist(testfile, headonly=True)
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 635
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        assert len(stream[0].data) == 0

    def test_read_slist_file_encoding(self, testdata):
        """
        Read SLIST file test via obspy.core.ascii._read_slist.
        """
        # float32
        testfile = testdata['slist_float.ascii']
        stream = _read_slist(testfile)
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 12
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        data = [185.01, 181.02, 185.03, 189.04, 194.05, 205.06,
                209.07, 214.08, 222.09, 225.98, 226.99, 219.00]
        np.testing.assert_array_almost_equal(stream[0].data, data, decimal=2)
        # unknown encoding
        testfile = testdata['slist_unknown.ascii']
        with pytest.raises(NotImplementedError):
            _read_slist(testfile)

    def test_is_tspair_file(self, testdata, datapath):
        """
        Testing TSPAIR file format.
        """
        testfile = testdata['tspair.ascii']
        assert _is_tspair(testfile)
        testfile = testdata['tspair_2_traces.ascii']
        assert _is_tspair(testfile)
        testfile = testdata['slist.ascii']
        assert not _is_tspair(testfile)
        # not existing file should fail
        testfile = datapath / 'xyz'
        assert not _is_tspair(testfile)

    def test_read_tspair_file_single_trace(self, testdata):
        """
        Read TSPAIR file test via obspy.core.ascii._read_tspair.
        """
        testfile = testdata['tspair.ascii']
        # read
        stream = _read_tspair(testfile)
        stream.verify()
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 635
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        assert stream[0].stats.mseed.dataquality == 'R'
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[0].data[0:4], data)
        # check last 4 samples
        data = [761, 755, 748, 746]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data)

    def test_read_tspair_file_multiple_traces(self, testdata):
        """
        Read TSPAIR file test via obspy.core.ascii._read_tspair.
        """
        testfile = testdata['tspair_2_traces.ascii']
        # read
        stream = _read_tspair(testfile)
        stream.verify()
        # sort traces to ensure comparable results
        stream.sort()
        assert stream[1].stats.network == 'XX'
        assert stream[1].stats.station == 'TEST'
        assert stream[1].stats.location == ''
        assert stream[1].stats.channel == 'BHZ'
        assert stream[1].stats.sampling_rate == 40.0
        assert stream[1].stats.npts == 635
        assert stream[1].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[1].stats.calib == 1.0e-00
        assert stream[1].stats.mseed.dataquality == 'R'
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[1].data[0:4], data)
        # check last 4 samples
        data = [761, 755, 748, 746]
        np.testing.assert_array_almost_equal(stream[1].data[-4:], data)
        # second trace
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHE'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 630
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        assert stream[0].stats.mseed.dataquality == 'R'
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[0].data[0:4], data)
        # check last 4 samples
        data = [781, 785, 778, 772]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data)

    def test_read_tspair_head_only(self, testdata):
        """
        Read TSPAIR file test via obspy.core.ascii._read_tspair.
        """
        testfile = testdata['tspair.ascii']
        # read
        stream = _read_tspair(testfile, headonly=True)
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 635
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        assert stream[0].stats.mseed.dataquality == 'R'
        assert len(stream[0].data) == 0

    def test_read_tspair_file_encoding(self, testdata):
        """
        Read TSPAIR file test via obspy.core.ascii._read_tspair.
        """
        # float32
        testfile = testdata['tspair_float.ascii']
        stream = _read_tspair(testfile)
        stream.verify()
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 12
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        assert stream[0].stats.mseed.dataquality == 'R'
        data = [185.01, 181.02, 185.03, 189.04, 194.05, 205.06,
                209.07, 214.08, 222.09, 225.98, 226.99, 219.00]
        np.testing.assert_array_almost_equal(stream[0].data, data, decimal=2)
        # unknown encoding
        testfile = testdata['tspair_unknown.ascii']
        with pytest.raises(NotImplementedError):
            _read_tspair(testfile)

    def test_write_tspair(self, testdata):
        """
        Write TSPAIR file test via obspy.core.ascii._write_tspair.
        """
        # float32
        testfile = testdata['tspair_float.ascii']
        stream_orig = _read_tspair(testfile)
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write
            _write_tspair(stream_orig, tmpfile)
            # look at the raw data
            with open(tmpfile, 'rt') as f:
                lines = f.readlines()
            assert lines[0].strip() == \
                'TIMESERIES XX_TEST__BHZ_R, 12 samples, 40 sps, ' + \
                '2008-01-15T00:00:00.025000, TSPAIR, FLOAT, Counts'
            assert lines[1].strip() == \
                '2008-01-15T00:00:00.025000  +1.8500999450e+02'
            # read again
            stream = _read_tspair(tmpfile)
            stream.verify()
            assert stream[0].stats.network == 'XX'
            assert stream[0].stats.station == 'TEST'
            assert stream[0].stats.location == ''
            assert stream[0].stats.channel == 'BHZ'
            assert stream[0].stats.sampling_rate == 40.0
            assert stream[0].stats.npts == 12
            assert stream[0].stats.starttime == \
                UTCDateTime("2008-01-15T00:00:00.025000")
            assert stream[0].stats.calib == 1.0e-00
            assert stream[0].stats.mseed.dataquality == 'R'
            data = [185.01, 181.02, 185.03, 189.04, 194.05, 205.06,
                    209.07, 214.08, 222.09, 225.98, 226.99, 219.00]
            np.testing.assert_array_almost_equal(stream[0].data, data,
                                                 decimal=2)
            # compare raw header
            with open(testfile, 'rt') as f:
                lines_orig = f.readlines()
            with open(tmpfile, 'rt') as f:
                lines_new = f.readlines()
        assert lines_orig[0] == lines_new[0]

    def test_write_tspair_custom_fmt(self, testdata):
        """
        Write TSPAIR file test via obspy.core.ascii._write_tspair.
        """
        # float
        testfile_orig = testdata['tspair_float.ascii']
        testfile = testdata['tspair_float_custom_fmt.ascii']
        stream_orig = _read_tspair(testfile_orig)
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write
            _write_tspair(stream_orig, tmpfile, custom_fmt='%3.14f')
            # look at the raw data
            with open(tmpfile, 'rt') as f:
                lines = f.readlines()
            assert lines[0].strip() == \
                'TIMESERIES XX_TEST__BHZ_R, 12 samples, 40 sps, ' + \
                '2008-01-15T00:00:00.025000, TSPAIR, FLOAT, Counts'
            assert lines[1].strip() == \
                '2008-01-15T00:00:00.025000  185.00999450000000'
            # read again
            stream = _read_tspair(tmpfile)
            stream.verify()
            assert stream[0].stats.network == 'XX'
            assert stream[0].stats.station == 'TEST'
            assert stream[0].stats.location == ''
            assert stream[0].stats.channel == 'BHZ'
            assert stream[0].stats.sampling_rate == 40.0
            assert stream[0].stats.npts == 12
            assert stream[0].stats.starttime == \
                UTCDateTime("2008-01-15T00:00:00.025000")
            assert stream[0].stats.calib == 1.0e-00
            assert stream[0].stats.mseed.dataquality == 'R'
            data = [185.01, 181.02, 185.03, 189.04, 194.05, 205.06,
                    209.07, 214.08, 222.09, 225.98, 226.99, 219.00]
            np.testing.assert_array_almost_equal(stream[0].data, data,
                                                 decimal=2)
            # compare raw header
            with open(testfile, 'rt') as f:
                lines_orig = f.readlines()
            with open(tmpfile, 'rt') as f:
                lines_new = f.readlines()
        assert lines_orig[0] == lines_new[0]

    def test_write_tspair_custom_fmt_custom(self, testdata):
        """
        Write TSPAIR file test via obspy.core.ascii._write_tspair.
        """
        # float
        testfile_orig = testdata['tspair_float.ascii']
        stream_orig = _read_tspair(testfile_orig)
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write
            _write_tspair(stream_orig, tmpfile, custom_fmt='%+r')
            with pytest.raises(NotImplementedError):
                _read_tspair(tmpfile)
            # look at the raw data
            with open(tmpfile, 'rt') as f:
                lines = f.readlines()
            assert lines[0].strip() == \
                'TIMESERIES XX_TEST__BHZ_R, 12 samples, 40 sps, ' + \
                '2008-01-15T00:00:00.025000, TSPAIR, CUSTOM, Counts'
            assert lines[1].strip() == \
                '2008-01-15T00:00:00.025000  185.0099945'

    def test_write_tspair_file_multiple_traces(self, testdata):
        """
        Write TSPAIR file test via obspy.core.ascii._write_tspair.
        """
        testfile = testdata['tspair_2_traces.ascii']
        stream_orig = _read_tspair(testfile)
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write
            _write_tspair(stream_orig, tmpfile)
            # look at the raw data
            with open(tmpfile, 'rt') as f:
                lines = f.readlines()
            assert lines[0].startswith('TIMESERIES')
            assert 'TSPAIR' in lines[0]
            assert lines[1] == '2008-01-15T00:00:00.025000  185\n'
            # test issue #321 (problems in time stamping)
            assert lines[-1] == '2008-01-15T00:00:15.750000  772\n'
            # read again
            stream = _read_tspair(tmpfile)
        stream.verify()
        # sort traces to ensure comparable results
        stream.sort()
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHE'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 630
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        assert stream[0].stats.mseed.dataquality == 'R'
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[0].data[0:4], data)
        # check last 4 samples
        data = [781, 785, 778, 772]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data)
        # second trace
        assert stream[1].stats.network == 'XX'
        assert stream[1].stats.station == 'TEST'
        assert stream[1].stats.location == ''
        assert stream[1].stats.channel == 'BHZ'
        assert stream[1].stats.sampling_rate == 40.0
        assert stream[1].stats.npts == 635
        assert stream[1].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[1].stats.calib == 1.0e-00
        assert stream[0].stats.mseed.dataquality == 'R'
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[1].data[0:4], data)
        # check last 4 samples
        data = [761, 755, 748, 746]
        np.testing.assert_array_almost_equal(stream[1].data[-4:], data)

    def test_write_slist(self, testdata):
        """
        Write SLIST file test via obspy.core.ascii._write_tspair.
        """
        # float32
        testfile = testdata['slist_float.ascii']
        stream_orig = _read_slist(testfile)
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write
            _write_slist(stream_orig, tmpfile)
            # look at the raw data
            with open(tmpfile, 'rt') as f:
                lines = f.readlines()
            assert lines[0].strip() == \
                'TIMESERIES XX_TEST__BHZ_R, 12 samples, 40 sps, ' + \
                '2008-01-15T00:00:00.025000, SLIST, FLOAT, Counts'
            assert lines[1].strip() == \
                '+1.8500999450e+02\t+1.8102000430e+02\t+1.8502999880e+02\t' + \
                '+1.8903999330e+02\t+1.9405000310e+02\t+2.0505999760e+02'
            # read again
            stream = _read_slist(tmpfile)
            stream.verify()
            assert stream[0].stats.network == 'XX'
            assert stream[0].stats.station == 'TEST'
            assert stream[0].stats.location == ''
            assert stream[0].stats.channel == 'BHZ'
            assert stream[0].stats.sampling_rate == 40.0
            assert stream[0].stats.npts == 12
            assert stream[0].stats.starttime == \
                UTCDateTime("2008-01-15T00:00:00.025000")
            assert stream[0].stats.calib == 1.0e-00
            assert stream[0].stats.mseed.dataquality == 'R'
            data = [185.01, 181.02, 185.03, 189.04, 194.05, 205.06,
                    209.07, 214.08, 222.09, 225.98, 226.99, 219.00]
            np.testing.assert_array_almost_equal(stream[0].data, data,
                                                 decimal=2)
            # compare raw header
            with open(testfile, 'rt') as f:
                lines_orig = f.readlines()
            with open(tmpfile, 'rt') as f:
                lines_new = f.readlines()
        assert lines_orig[0] == lines_new[0]

    def test_write_slist_custom_fmt_float(self, testdata):
        """
        Write SLIST file test via obspy.core.ascii._write_tspair.
        """
        # float
        testfile_orig = testdata['slist_float.ascii']
        testfile = testdata['slist_float_custom_fmt.ascii']
        stream_orig = _read_slist(testfile_orig)
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write
            _write_slist(stream_orig, tmpfile, custom_fmt='%3.14f')
            # look at the raw data
            with open(tmpfile, 'rt') as f:
                lines = f.readlines()
            assert lines[0].strip() == \
                'TIMESERIES XX_TEST__BHZ_R, 12 samples, 40 sps, ' + \
                '2008-01-15T00:00:00.025000, SLIST, FLOAT, Counts'
            assert lines[1].strip() == \
                '185.00999450000000\t181.02000430000001\t' + \
                '185.02999879999999\t189.03999329999999\t' + \
                '194.05000310000000\t205.05999760000000'
            # read again
            stream = _read_slist(tmpfile)
            stream.verify()
            assert stream[0].stats.network == 'XX'
            assert stream[0].stats.station == 'TEST'
            assert stream[0].stats.location == ''
            assert stream[0].stats.channel == 'BHZ'
            assert stream[0].stats.sampling_rate == 40.0
            assert stream[0].stats.npts == 12
            assert stream[0].stats.starttime == \
                UTCDateTime("2008-01-15T00:00:00.025000")
            assert stream[0].stats.calib == 1.0e-00
            assert stream[0].stats.mseed.dataquality == 'R'
            data = [185.01, 181.02, 185.03, 189.04, 194.05, 205.06,
                    209.07, 214.08, 222.09, 225.98, 226.99, 219.00]
            np.testing.assert_array_almost_equal(stream[0].data, data,
                                                 decimal=2)
            # compare raw header
            with open(testfile, 'rt') as f:
                lines_orig = f.readlines()
            with open(tmpfile, 'rt') as f:
                lines_new = f.readlines()
        assert lines_orig[0] == lines_new[0]

    def test_write_slist_custom_fmt_custom(self, testdata):
        """
        Write SLIST file test via obspy.core.ascii._write_tspair.
        """
        # float
        testfile_orig = testdata['slist_float.ascii']
        stream_orig = _read_slist(testfile_orig)
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write
            _write_slist(stream_orig, tmpfile, custom_fmt='%+r')
            with pytest.raises(NotImplementedError):
                _read_slist(tmpfile)
            # look at the raw data
            with open(tmpfile, 'rt') as f:
                lines = f.readlines()
            assert lines[0].strip() == \
                'TIMESERIES XX_TEST__BHZ_R, 12 samples, 40 sps, ' + \
                '2008-01-15T00:00:00.025000, SLIST, CUSTOM, Counts'
            assert lines[1].strip() == \
                '185.0099945\t181.02000430000001\t' + \
                '185.02999879999999\t189.03999329999999\t' + \
                '194.0500031\t205.0599976'

    def test_write_slist_file_multiple_traces(self, testdata):
        """
        Write SLIST file test via obspy.core.ascii._write_tspair.
        """
        testfile = testdata['slist_2_traces.ascii']
        stream_orig = _read_slist(testfile)
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # write
            _write_slist(stream_orig, tmpfile)
            # look at the raw data
            with open(tmpfile, 'rt') as f:
                lines = f.readlines()
            assert lines[0].startswith('TIMESERIES')
            assert 'SLIST' in lines[0]
            assert lines[1].strip() == '185\t181\t185\t189\t194\t205'
            # read again
            stream = _read_slist(tmpfile)
        stream.verify()
        # sort traces to ensure comparable results
        stream.sort()
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHE'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 630
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        assert stream[0].stats.mseed.dataquality == 'R'
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[0].data[0:4], data)
        # check last 4 samples
        data = [781, 785, 778, 772]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data)
        # second trace
        assert stream[1].stats.network == 'XX'
        assert stream[1].stats.station == 'TEST'
        assert stream[1].stats.location == ''
        assert stream[1].stats.channel == 'BHZ'
        assert stream[1].stats.sampling_rate == 40.0
        assert stream[1].stats.npts == 635
        assert stream[1].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[1].stats.calib == 1.0e-00
        assert stream[0].stats.mseed.dataquality == 'R'
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[1].data[0:4], data)
        # check last 4 samples
        data = [761, 755, 748, 746]
        np.testing.assert_array_almost_equal(stream[1].data[-4:], data)

    def test_write_small_trace(self):
        """
        Tests writing Traces containing 0, 1 or 2 samples only.
        """
        for format in ['SLIST', 'TSPAIR']:
            for num in range(0, 4):
                tr = Trace(data=np.arange(num))
                with NamedTemporaryFile() as tf:
                    tempfile = tf.name
                    tr.write(tempfile, format=format)
                    # test results
                    st = read(tempfile, format=format)
                assert len(st) == 1
                assert len(st[0]) == num

    def test_float_sampling_rates_write_and_read(self):
        """
        Tests writing and reading Traces with floating point and with less than
        1 Hz sampling rates.
        """
        tr = Trace(np.arange(10))
        check_sampling_rates = (0.000000001, 1.000000001, 100.000000001,
                                99.999999999, 1.5, 1.666666, 10000.0001)
        for format in ['SLIST', 'TSPAIR']:
            for sps in check_sampling_rates:
                tr.stats.sampling_rate = sps
                with NamedTemporaryFile() as tf:
                    tempfile = tf.name
                    tr.write(tempfile, format=format)
                    # test results
                    got = read(tempfile, format=format)[0]
                assert tr.stats.sampling_rate == got.stats.sampling_rate

    def test_determine_dtype(self):
        """
        Tests _determine_dtype for properly returned types
        """
        float_formats = ['%+10.10e', '%+.10e', '%.3e',
                         '%+10.10E', '%+.10E', '%.3E',
                         '%+10.10f', '%+.10f', '%.3f',
                         '%+10.10F', '%+.10F', '%.3F',
                         '%+10.10g', '%+.10g', '%.3g',
                         '%+10.10G', '%+.10G', '%.3G']

        int_formats = ['%+10.10i', '%+.10i', '%.3i',
                       '%+10.10I', '%+.10I', '%.3I',
                       '%+10.10d', '%+.10d', '%.3d',
                       '%+10.10D', '%+.10D', '%.3D']

        custom_formats = ['%+10.10s', '%+.10s', '%.3s',
                          '%+10.10x', '%+.10x', '%.3x',
                          '%+10.10k', '%+.10k', '%.3k',
                          '%+10.10z', '%+.10z', '%.3z',
                          '%+10.10w', '%+.10w', '%.3w',
                          '%+10.10q', '%+.10q', '%.3q']

        for format in float_formats:
            assert 'FLOAT' == _determine_dtype(format)

        for format in int_formats:
            assert 'INTEGER' == _determine_dtype(format)

        for format in custom_formats:
            assert 'CUSTOM' == _determine_dtype(format)

        with pytest.raises(ValueError):
            _determine_dtype('')

    def test_regression_against_mseed2ascii(self, testdata):
        """
        Regression test against issue #2165.
        """
        mseed_file = testdata["miniseed_record.mseed"]
        mseed2ascii_file = testdata["mseed2ascii_miniseed_record.txt"]

        with NamedTemporaryFile() as tf:
            # Write as TSPAIR
            read(mseed_file).write(tf.name, format="TSPAIR")
            # Check all lines aside from the first as they differ.
            with open(tf.name, "rt") as fh:
                actual_lines = fh.readlines()[1:]
            with open(mseed2ascii_file, "rt") as fh:
                expected_lines = fh.readlines()[1:]

        for actual, expected in zip(actual_lines, expected_lines):
            assert actual.strip() == expected.strip()

    def test_read_slist_file_single_trace_int32(self, testdata):
        """
        Read SLIST file test via obspy.core.ascii._read_slist.
        """
        testfile = testdata['slist_int32.ascii']
        # read
        stream = _read_slist(testfile)
        stream.verify()
        assert stream[0].stats.network == 'XX'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.location == ''
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.sampling_rate == 40.0
        assert stream[0].stats.npts == 635
        assert stream[0].stats.starttime == \
            UTCDateTime("2008-01-15T00:00:00.025000")
        assert stream[0].stats.calib == 1.0e-00
        assert stream[0].data.dtype.name == "int32"
        # check first 4 samples
        data = [185, 181, 185, 189]
        np.testing.assert_array_almost_equal(stream[0].data[0:4], data)
        # check last 4 samples
        data = [761, 755, 748, 746]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data)

    def test_read_slist_file_single_trace_float64(self, testdata):
        """
        Read SLIST file test via obspy.core.ascii._read_slist.
        """
        testfile = testdata['slist_float32.ascii']
        # read
        stream = _read_slist(testfile)
        stream.verify()
        assert stream[0].data.dtype.name == "float32"



# -*- coding: utf-8 -*-
import numpy as np
import pytest

from obspy.core.inventory import (Response, PolesZerosResponseStage,
                                  ResponseStage, CoefficientsTypeResponseStage)
from obspy.clients.nrl.client import NRL, LocalNRL, RemoteNRL


@pytest.mark.network
class TestNRLRemote():
    """
    Minimal NRL test suite connecting to online NRL

    """
    def test_nrl_type(self):
        nrl_online = NRL(root='http://ds.iris.edu/NRL')
        assert isinstance(nrl_online, RemoteNRL)


class TestNRLLocal():
    """
    NRL test suite for test cases common to v1 and v2 of NRL without network
    usage.
    """
    def test_error_handling_invalid_path(self):
        msg = \
            "Provided path '/some/really/random/path' seems to be a local " \
            "file path but the directory does not exist."
        with pytest.raises(ValueError, match=msg):
            NRL("/some/really/random/path")


class TestNRLLocalV1():
    """
    NRL test suite using stripped down local NRL without network usage.

    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, datapath):
        # Longer diffs in the test assertions.
        self.maxDiff = None
        # Small subset of NRL included in tests/data
        self.local_dl_key = ['REF TEK', 'RT 130 & 130-SMA', '1', '1']
        self.local_sensor_key = ['Guralp', 'CMG-3T', '120s - 50Hz', '1500']
        self.local_nrl_root = str(datapath / 'IRIS')
        self.nrl_local = NRL(root=self.local_nrl_root)

    def test_nrl_type(self):
        assert isinstance(self.nrl_local, LocalNRL)
        assert self.nrl_local._nrl_version == 1

    def test_get_response(self):
        # Get only the sensor response.
        sensor_resp = self.nrl_local.get_sensor_response(self.local_sensor_key)

        # Get only the datalogger response.
        dl_resp = self.nrl_local.get_datalogger_response(self.local_dl_key)

        # Get full response.
        resp = self.nrl_local.get_response(
            datalogger_keys=self.local_dl_key,
            sensor_keys=self.local_sensor_key)

        # Make sure that NRL.get_response() has overall instrument sensitivity
        # correctly recalculated after combining sensor and datalogger
        # information, see #3099.
        # Before fixing this bug the result was 945089653.7285056 which is a
        # relative deviation of 0.00104
        assert resp.instrument_sensitivity.value == pytest.approx(
            944098418.0614196, abs=0, rel=1e-4)

        # All of them should be Response objects.
        assert isinstance(resp, Response)
        assert isinstance(dl_resp, Response)
        assert isinstance(sensor_resp, Response)

        # The full response is the first stage from the sensor and all
        # following from the datalogger.
        assert resp.response_stages[0] == sensor_resp.response_stages[0]
        assert resp.response_stages[1:] == dl_resp.response_stages[1:]

        # Test the actual responses. Testing the parsing of the exact values
        # and what not is done in obspy.io.xseed.
        paz = sensor_resp.response_stages[0]
        assert isinstance(paz, PolesZerosResponseStage)
        np.testing.assert_allclose(
            paz.poles, [(-0.037008 + 0.037008j), (-0.037008 - 0.037008j),
                        (-502.65 + 0j), (-1005 + 0j), (-1131 + 0j)])
        np.testing.assert_allclose(paz.zeros, [0j, 0j])

        assert len(dl_resp.response_stages) == 15
        assert len(resp.response_stages) == 15

        assert isinstance(resp.response_stages[1], ResponseStage)
        for _i in range(2, 15):
            assert isinstance(resp.response_stages[_i],
                              CoefficientsTypeResponseStage)

    def test_nrl_class_str_method(self):
        out = str(self.nrl_local)
        # The local NRL is not going to chance so it is fine to test this.
        assert out.strip() == """
NRL library at %s
  Sensors: 20 manufacturers
    'CEA-DASE', 'CME', 'Chaparral Physics', 'Eentec', 'Generic',
    'Geo Space/OYO', 'Geodevice', 'Geotech', 'Guralp', 'Hyperion',
    'IESE', 'Kinemetrics', 'Lennartz', 'Metrozet', 'Nanometrics',
    'REF TEK', 'Sercel/Mark Products', 'SolGeo',
    'Sprengnether (now Eentec)', 'Streckeisen'
  Dataloggers: 13 manufacturers
    'Agecodagis', 'DAQ Systems (NetDAS)', 'Earth Data', 'Eentec',
    'Generic', 'Geodevice', 'Geotech', 'Guralp', 'Kinemetrics',
    'Nanometrics', 'Quanterra', 'REF TEK', 'SolGeo'
        """.strip() % self.local_nrl_root

    def test_nrl_dict_str_method(self):
        out = str(self.nrl_local.sensors)
        assert out.strip() == """
Select the sensor manufacturer (20 items):
  'CEA-DASE', 'CME', 'Chaparral Physics', 'Eentec', 'Generic',
  'Geo Space/OYO', 'Geodevice', 'Geotech', 'Guralp', 'Hyperion',
  'IESE', 'Kinemetrics', 'Lennartz', 'Metrozet', 'Nanometrics',
  'REF TEK', 'Sercel/Mark Products', 'SolGeo',
  'Sprengnether (now Eentec)', 'Streckeisen'""".strip()

    def test_error_handling_invalid_path(self):
        msg = \
            "Provided path '/some/really/random/path' seems to be a local " \
            "file path but the directory does not exist."
        with pytest.raises(ValueError, match=msg):
            NRL("/some/really/random/path")


class TestNRLLocalV2RESP():
    """
    NRL test suite using stripped down local NRL version 2 in RESP format
    without network usage.
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, datapath):
        # Longer diffs in the test assertions.
        self.maxDiff = None
        # Small subset of NRL included in tests/data
        self.local_dl_key = ['REFTEK', '130-01', '1', '1 Hz']
        self.local_sensor_key = ['Guralp', 'CMG-3T', '120 s', '50 Hz', '1500']
        self.local_nrl_root = str(datapath / 'IRIS_v2_resp')
        self.nrl_local = NRL(root=self.local_nrl_root)

    def test_nrl_type(self):
        assert isinstance(self.nrl_local, LocalNRL)
        assert self.nrl_local._nrl_version == 2

    def test_get_response(self):
        # Get only the sensor response.
        sensor_resp = self.nrl_local.get_sensor_response(self.local_sensor_key)

        # Get only the datalogger response.
        msg = r"First stage input units are 'COUNTS'\. When requesting a .*"
        with pytest.warns(UserWarning, match=msg):
            dl_resp = self.nrl_local.get_datalogger_response(self.local_dl_key)

        # Get full response.
        resp = self.nrl_local.get_response(
            datalogger_keys=self.local_dl_key,
            sensor_keys=self.local_sensor_key)

        # Make sure that NRL.get_response() has overall instrument sensitivity
        # correctly recalculated after combining sensor and datalogger
        # information, see #3099.
        # Before fixing this bug the result was 945089653.7285056 which is a
        # relative deviation of 0.00104
        assert resp.instrument_sensitivity.value == pytest.approx(
            945084144.2013303, abs=0, rel=1e-4)

        # All of them should be Response objects.
        assert isinstance(resp, Response)
        assert isinstance(dl_resp, Response)
        assert isinstance(sensor_resp, Response)

        # Changed in NRL v2, the full response is the all stages from the
        # sensor followed by all response stages from the datalogger, so there
        # is no dummy stage 1 anymore in the datalogger RESP files.
        # That means that the stage sequence numbers get changed during
        # assembling the full response and they have to be adjusted here
        # accordingly
        num_sensor_stages = len(sensor_resp.response_stages)
        # we also fix the missing units of first datalogger stage when a
        # combined sensor+datalogger response is requested, which doesn't work
        # when fetching the datalogger-only response, so we need to adjust
        # these units here for the test manually
        dl_resp.response_stages[0].input_units = 'V'
        dl_resp.response_stages[0].input_units_description = 'Volts'
        dl_resp.response_stages[0].output_units_description = 'Volts'

        for stage in dl_resp.response_stages:
            stage.stage_sequence_number += num_sensor_stages
        assert resp.response_stages[:num_sensor_stages] == \
            sensor_resp.response_stages[:]
        assert resp.response_stages[num_sensor_stages:] == \
            dl_resp.response_stages[:]

        # Test the actual responses. Testing the parsing of the exact values
        # and what not is done in obspy.io.xseed.
        paz = sensor_resp.response_stages[0]
        assert isinstance(paz, PolesZerosResponseStage)
        np.testing.assert_allclose(
            paz.poles, [(-0.037008 - 0.037008j), (-0.037008 + 0.037008j),
                        (-502.65 + 0j), (-1005 + 0j), (-1131 + 0j)])
        np.testing.assert_allclose(paz.zeros, [0j, 0j])

        assert len(dl_resp.response_stages) == 14
        assert len(resp.response_stages) == 15

        assert isinstance(resp.response_stages[1], ResponseStage)
        for _i in range(2, 15):
            assert isinstance(resp.response_stages[_i],
                              CoefficientsTypeResponseStage)

    def test_nrl_class_str_method(self):
        out = str(self.nrl_local)
        # The local NRL is not going to chance so it is fine to test this.
        assert out.strip() == """
NRL library at %s
  Sensors: 38 manufacturers
    'ASIR', 'CEADASE', 'Chaparral', 'DTCC', 'EQMet', 'Eentec',
    'GEObit', 'GaiaCode', 'GenericUnity', 'GeoDevice', 'GeoSIG',
    'GeoSpace', 'Geotech', 'Guralp', 'HGSProducts', 'HighTech',
    'Hyperion', 'IESE', 'JeffreyBJohnson', 'Kinemetrics', 'LaHusen',
    'Lennartz', 'Lunitek', 'MagseisFairfield', 'Nanometrics', 'REFTEK'
    'RSensors', 'RTClark', 'SARA', 'SeismoWave', 'SensorNederland',
    'Sercel', 'SiliconAudio', 'SolGeo', 'Sprengnether', 'Streckeisen'
    'Sunfull', 'iTem'
  Dataloggers: 28 manufacturers
    'Agecodagis', 'CNSN', 'DAQSystems', 'DTCC', 'DiGOSOmnirecs',
    'EQMet', 'EarthData', 'Eentec', 'GEObit', 'GenericUnity',
    'GeoDevice', 'GeoSIG', 'Geotech', 'Guralp', 'Kinemetrics',
    'Lunitek', 'MagseisFairfield', 'Nanometrics', 'Quanterra',
    'REFTEK', 'RSensors', 'SARA', 'STANEO', 'SeismicSource',
    'SeismologyResearchCentre', 'Sercel', 'SolGeo', 'WorldSensing'
        """.strip() % self.local_nrl_root

    def test_nrl_dict_str_method(self):
        out = str(self.nrl_local.sensors)
        assert out.strip() == """
Select the sensor manufacturer (38 items):
  'ASIR', 'CEADASE', 'Chaparral', 'DTCC', 'EQMet', 'Eentec', 'GEObit'
  'GaiaCode', 'GenericUnity', 'GeoDevice', 'GeoSIG', 'GeoSpace',
  'Geotech', 'Guralp', 'HGSProducts', 'HighTech', 'Hyperion', 'IESE',
  'JeffreyBJohnson', 'Kinemetrics', 'LaHusen', 'Lennartz', 'Lunitek',
  'MagseisFairfield', 'Nanometrics', 'REFTEK', 'RSensors', 'RTClark',
  'SARA', 'SeismoWave', 'SensorNederland', 'Sercel', 'SiliconAudio',
  'SolGeo', 'Sprengnether', 'Streckeisen', 'Sunfull', 'iTem'""".strip()


class TestNRLLocalV2StationXML():
    """
    NRL test suite using stripped down local NRL version 2 in StationXML format
    without network usage.
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, datapath):
        # Longer diffs in the test assertions.
        self.maxDiff = None
        # Small subset of NRL included in tests/data
        self.local_dl_key = ['REFTEK', '130-01', '1', '1 Hz']
        self.local_sensor_key = ['Guralp', 'CMG-3T', '120 s', '50 Hz', '1500']
        self.local_nrl_root = str(datapath / 'IRIS_v2_stationxml')
        self.nrl_local = NRL(root=self.local_nrl_root)

    def test_nrl_type(self):
        assert isinstance(self.nrl_local, LocalNRL)
        assert self.nrl_local._nrl_version == 2

    def test_get_response(self):
        # Get only the sensor response.
        sensor_resp = self.nrl_local.get_sensor_response(self.local_sensor_key)

        # Get only the datalogger response.
        msg = r"Undefined input units in stage one\. Most datalogger-only.*"
        with pytest.warns(UserWarning, match=msg):
            dl_resp = self.nrl_local.get_datalogger_response(self.local_dl_key)

        # Get full response.
        resp = self.nrl_local.get_response(
            datalogger_keys=self.local_dl_key,
            sensor_keys=self.local_sensor_key)

        # Make sure that NRL.get_response() has overall instrument sensitivity
        # correctly recalculated after combining sensor and datalogger
        # information, see #3099.
        # Before fixing this bug the result was 945089653.7285056 which is a
        # relative deviation of 0.00104
        assert resp.instrument_sensitivity.value == pytest.approx(
            945084144.2013303, abs=0, rel=1e-4)

        # All of them should be Response objects.
        assert isinstance(resp, Response)
        assert isinstance(dl_resp, Response)
        assert isinstance(sensor_resp, Response)

        # Changed in NRL v2, the full response is the all stages from the
        # sensor followed by all response stages from the datalogger, so there
        # is no dummy stage 1 anymore in the datalogger RESP files.
        # That means that the stage sequence numbers get changed during
        # assembling the full response and they have to be adjusted here
        # accordingly
        num_sensor_stages = len(sensor_resp.response_stages)
        for stage in dl_resp.response_stages:
            stage.stage_sequence_number += num_sensor_stages
        # we also fix the missing units of first datalogger stage when a
        # combined sensor+datalogger response is requested, which doesn't work
        # when fetching the datalogger-only response, so we need to adjust
        # these units here for the test manually
        dl_resp.response_stages[0].input_units = 'V'
        dl_resp.response_stages[0].input_units_description = 'Volts'
        # dl_resp.response_stages[0].output_units_description = 'Volts'

        assert resp.response_stages[:num_sensor_stages] == \
            sensor_resp.response_stages[:]
        assert resp.response_stages[num_sensor_stages:] == \
            dl_resp.response_stages[:]

        # Test the actual responses. Testing the parsing of the exact values
        # and what not is done in obspy.io.xseed.
        paz = sensor_resp.response_stages[0]
        assert isinstance(paz, PolesZerosResponseStage)
        np.testing.assert_allclose(
            paz.poles, [(-0.037008 - 0.037008j), (-0.037008 + 0.037008j),
                        (-502.65 + 0j), (-1005 + 0j), (-1131 + 0j)])
        np.testing.assert_allclose(paz.zeros, [0j, 0j])

        assert len(dl_resp.response_stages) == 14
        assert len(resp.response_stages) == 15

        assert isinstance(resp.response_stages[1], ResponseStage)
        for _i in range(2, 15):
            assert isinstance(resp.response_stages[_i],
                              CoefficientsTypeResponseStage)

    def test_nrl_class_str_method(self):
        out = str(self.nrl_local)
        # The local NRL is not going to chance so it is fine to test this.
        assert out.strip() == """
NRL library at %s
  Sensors: 38 manufacturers
    'ASIR', 'CEADASE', 'Chaparral', 'DTCC', 'EQMet', 'Eentec',
    'GEObit', 'GaiaCode', 'GenericUnity', 'GeoDevice', 'GeoSIG',
    'GeoSpace', 'Geotech', 'Guralp', 'HGSProducts', 'HighTech',
    'Hyperion', 'IESE', 'JeffreyBJohnson', 'Kinemetrics', 'LaHusen',
    'Lennartz', 'Lunitek', 'MagseisFairfield', 'Nanometrics', 'REFTEK'
    'RSensors', 'RTClark', 'SARA', 'SeismoWave', 'SensorNederland',
    'Sercel', 'SiliconAudio', 'SolGeo', 'Sprengnether', 'Streckeisen'
    'Sunfull', 'iTem'
  Dataloggers: 28 manufacturers
    'Agecodagis', 'CNSN', 'DAQSystems', 'DTCC', 'DiGOSOmnirecs',
    'EQMet', 'EarthData', 'Eentec', 'GEObit', 'GenericUnity',
    'GeoDevice', 'GeoSIG', 'Geotech', 'Guralp', 'Kinemetrics',
    'Lunitek', 'MagseisFairfield', 'Nanometrics', 'Quanterra',
    'REFTEK', 'RSensors', 'SARA', 'STANEO', 'SeismicSource',
    'SeismologyResearchCentre', 'Sercel', 'SolGeo', 'WorldSensing'
        """.strip() % self.local_nrl_root

    def test_nrl_dict_str_method(self):
        out = str(self.nrl_local.sensors)
        assert out.strip() == """
Select the sensor manufacturer (38 items):
  'ASIR', 'CEADASE', 'Chaparral', 'DTCC', 'EQMet', 'Eentec', 'GEObit'
  'GaiaCode', 'GenericUnity', 'GeoDevice', 'GeoSIG', 'GeoSpace',
  'Geotech', 'Guralp', 'HGSProducts', 'HighTech', 'Hyperion', 'IESE',
  'JeffreyBJohnson', 'Kinemetrics', 'LaHusen', 'Lennartz', 'Lunitek',
  'MagseisFairfield', 'Nanometrics', 'REFTEK', 'RSensors', 'RTClark',
  'SARA', 'SeismoWave', 'SensorNederland', 'Sercel', 'SiliconAudio',
  'SolGeo', 'Sprengnether', 'Streckeisen', 'Sunfull', 'iTem'""".strip()

    def test_get_integrated_response(self):
        # Get an integrated response
        resp = self.nrl_local.get_integrated_response(
            ['GeoSIG', 'ela-GMSseries', '0 Hz', '1000 Hz', '100 Hz', '1'])

        assert len(resp.response_stages) == 5
        assert resp.instrument_sensitivity.value == 778678
        assert resp.instrument_sensitivity.input_units == 'm/s**2'
        assert resp.instrument_sensitivity.output_units == 'counts'

    def test_get_soh_response(self):
        # Get an integrated response
        resp = self.nrl_local.get_soh_response(
            ['Quanterra', 'Q330PacketBaler', 'MassPosition', '10', '0.1 Hz'])

        assert len(resp.response_stages) == 1
        assert resp.instrument_sensitivity.value == 0.0
        assert resp.instrument_sensitivity.input_units == 'V'
        assert resp.instrument_sensitivity.output_units == 'counts'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test suite for the network class.
:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2013
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import io
import warnings
from unittest import mock

import numpy as np
import pytest

import obspy
from obspy import UTCDateTime, read_inventory
from obspy.core.inventory import (Channel, Inventory, Network, Response,
                                  Station)
from obspy.core.util import CARTOPY_VERSION
from obspy.core.util.base import CatchAndAssertWarnings


@pytest.mark.usefixtures('ignore_numpy_errors')
class TestNetwork:
    """
    Tests for the :class:`~obspy.core.inventory.network.Network` class.
    """
    def test_get_response(self):
        response_n1_s1 = Response('RESPN1S1')
        response_n1_s2 = Response('RESPN1S2')
        response_n2_s1 = Response('RESPN2S1')
        channels_n1_s1 = [Channel(code='BHZ',
                                  location_code='',
                                  latitude=0.0,
                                  longitude=0.0,
                                  elevation=0.0,
                                  depth=0.0,
                                  response=response_n1_s1)]
        channels_n1_s2 = [Channel(code='BHZ',
                                  location_code='',
                                  latitude=0.0,
                                  longitude=0.0,
                                  elevation=0.0,
                                  depth=0.0,
                                  response=response_n1_s2)]
        channels_n2_s1 = [Channel(code='BHZ',
                                  location_code='',
                                  latitude=0.0,
                                  longitude=0.0,
                                  elevation=0.0,
                                  depth=0.0,
                                  response=response_n2_s1)]
        stations_1 = [Station(code='N1S1',
                              latitude=0.0,
                              longitude=0.0,
                              elevation=0.0,
                              channels=channels_n1_s1),
                      Station(code='N1S2',
                              latitude=0.0,
                              longitude=0.0,
                              elevation=0.0,
                              channels=channels_n1_s2),
                      Station(code='N2S1',
                              latitude=0.0,
                              longitude=0.0,
                              elevation=0.0,
                              channels=channels_n2_s1)]
        network = Network('N1', stations=stations_1)

        response = network.get_response('N1.N1S1..BHZ',
                                        UTCDateTime('2010-01-01T12:00'))
        assert response == response_n1_s1
        response = network.get_response('N1.N1S2..BHZ',
                                        UTCDateTime('2010-01-01T12:00'))
        assert response == response_n1_s2
        response = network.get_response('N1.N2S1..BHZ',
                                        UTCDateTime('2010-01-01T12:00'))
        assert response == response_n2_s1

    def test_get_coordinates(self):
        """
        Test extracting coordinates
        """
        expected = {u'latitude': 47.737166999999999,
                    u'longitude': 12.795714,
                    u'elevation': 860.0,
                    u'local_depth': 0.0}
        channels = [Channel(code='EHZ',
                            location_code='',
                            start_date=UTCDateTime('2007-01-01'),
                            latitude=47.737166999999999,
                            longitude=12.795714,
                            elevation=860.0,
                            depth=0.0)]
        stations = [Station(code='RJOB',
                            latitude=0.0,
                            longitude=0.0,
                            elevation=0.0,
                            channels=channels)]
        network = Network('BW', stations=stations)
        # 1
        coordinates = network.get_coordinates('BW.RJOB..EHZ',
                                              UTCDateTime('2010-01-01T12:00'))
        assert sorted(coordinates.items()) == sorted(expected.items())
        # 2 - without datetime
        coordinates = network.get_coordinates('BW.RJOB..EHZ')
        assert sorted(coordinates.items()) == sorted(expected.items())
        # 3 - unknown SEED ID should raise exception
        with pytest.raises(Exception):
            network.get_coordinates('BW.RJOB..XXX')

    def test_response_plot(self, image_path):
        """
        Tests the response plot.
        """
        net = read_inventory()[0]
        t = UTCDateTime(2008, 7, 1)
        with CatchAndAssertWarnings():
            net.plot_response(0.002, output="DISP", channel="B*E",
                              time=t, outfile=image_path)

    def test_response_plot_epoch_times_in_label(self):
        """
        Tests response plot with epoch times in labels switched on.
        """
        import matplotlib.pyplot as plt
        net = read_inventory().select(station='RJOB', channel='EHZ')[0]
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("ignore")
            fig = net.plot_response(0.01, label_epoch_dates=True, show=False)
        try:
            legend = fig.axes[0].get_legend()
            texts = legend.get_texts()
            expecteds = ['BW.RJOB..EHZ\n2001-05-15 -- 2006-12-12',
                         'BW.RJOB..EHZ\n2006-12-13 -- 2007-12-17',
                         'BW.RJOB..EHZ\n2007-12-17 -- open']
            assert len(texts) == 3
            for text, expected in zip(texts, expecteds):
                assert text.get_text() == expected
        finally:
            plt.close(fig)

    def test_len(self):
        """
        Tests the __len__ property.
        """
        net = read_inventory()[0]
        assert len(net) == len(net.stations)
        assert len(net) == 2

    def test_network_select(self):
        """
        Test for the select() method of the network class.
        """
        net = read_inventory()[0]

        # Basic asserts to assert some things about the test data.
        assert len(net) == 2
        assert len(net[0]) == 12
        assert len(net[1]) == 9
        assert sum(len(i) for i in net) == 21

        # Artificially move the start time of the first station before the
        # channel start times.
        net[0].start_date = UTCDateTime(1999, 1, 1)

        # Nothing happens if nothing is specified or if everything is selected.
        assert sum(len(i) for i in net.select()) == 21
        assert sum(len(i) for i in net.select(station="*")) == 21
        assert sum(len(i) for i in net.select(location="*")) == 21
        assert sum(len(i) for i in net.select(channel="*")) == 21
        subsum = sum(
            len(i) for i in net.select(station="*", location="*", channel="*")
        )
        assert subsum == 21

        # No matching station.
        assert sum(len(i) for i in net.select(station="RR")) == 0
        # keep_empty does not do anything in these cases.
        sub = sum(len(i) for i in net.select(station="RR", keep_empty=True))
        assert sub == 0
        # Selecting only one station.
        sub = sum(len(i) for i in net.select(station="FUR", keep_empty=True))
        assert sub == 12
        sub = sum(len(i) for i in net.select(station="F*", keep_empty=True))
        assert sub == 12
        sub = sum(len(i) for i in net.select(station="WET", keep_empty=True))
        assert sub == 9
        kwargs = dict(
            minlatitude=47.89, maxlatitude=48.39,
            minlongitude=10.88, maxlongitude=11.98,
        )
        sub = sum((len(i) for i in net.select(**kwargs)))
        assert sub == 12
        sub = sum(
            len(i) for i in net.select(
                latitude=48.12, longitude=12.24, maxradius=1)
        )
        assert sub == 12

        # Test the keep_empty flag.
        net_2 = net.select(time=UTCDateTime(2000, 1, 1))
        assert len(net_2) == 0
        assert sum(len(i) for i in net_2) == 0
        # One is kept - it has no more channels but the station still has a
        # valid start time.
        net_2 = net.select(time=UTCDateTime(2000, 1, 1), keep_empty=True)
        assert len(net_2) == 1
        assert sum(len(i) for i in net_2) == 0

        # location, channel, time, starttime, endtime, and sampling_rate
        # and geographic parameters are also passed on to the station selector.
        select_kwargs = {
            "location": "00",
            "channel": "EHE",
            "time": UTCDateTime(2001, 1, 1),
            "sampling_rate": 123.0,
            "starttime": UTCDateTime(2002, 1, 1),
            "endtime": UTCDateTime(2003, 1, 1),
            "minlatitude": None,
            "maxlatitude": None,
            "minlongitude": None,
            "maxlongitude": None,
            "latitude": None,
            "longitude": None,
            "minradius": None,
            "maxradius": None}

        with mock.patch("obspy.core.inventory.station.Station.select") as p:
            p.return_value = obspy.core.inventory.station.Station(
                "FUR", 1, 2, 3,
            )
            net.select(**select_kwargs)

        assert p.call_args[1] == select_kwargs

    def test_writing_network_before_1990(self):
        inv = obspy.Inventory(networks=[
            Network(code="XX", start_date=obspy.UTCDateTime(1880, 1, 1))],
            source="")
        with io.BytesIO() as buf:
            inv.write(buf, format="stationxml")
            buf.seek(0, 0)
            inv2 = read_inventory(buf)

        assert inv.networks[0] == inv2.networks[0]

    def test_network_select_with_empty_stations(self):
        """
        Tests the behaviour of the Network.select() method for empty stations.
        """
        net = read_inventory()[0]

        # Delete all channels.
        for sta in net:
            sta.channels = []

        # 2 stations and 0 channels remain.
        assert len(net) == 2
        assert sum(len(sta) for sta in net) == 0

        # No arguments, everything should be selected.
        assert len(net.select()) == 2

        # Everything selected, nothing should happen.
        assert len(net.select(station="*")) == 2

        # Only select a single station.
        assert len(net.select(station="FUR")) == 1
        assert len(net.select(station="FU?")) == 1
        assert len(net.select(station="W?T")) == 1

        # Once again, this time with the time selection.
        assert len(net.select(time=UTCDateTime(2006, 1, 1))) == 0
        assert len(net.select(time=UTCDateTime(2007, 1, 1))) == 1
        assert len(net.select(time=UTCDateTime(2008, 1, 1))) == 2

    def test_empty_network_code(self):
        """
        Tests that an empty sring is acceptabble.
        """
        # An empty string is allowed.
        n = Network(code="")
        assert n.code == ""

        # But None is not allowed.
        with pytest.raises(ValueError, match='A code is required'):
            Network(code=None)

        # Should still serialize to something.
        inv = Inventory(networks=[n])
        with io.BytesIO() as buf:
            inv.write(buf, format="stationxml", validate=True)
            buf.seek(0, 0)
            inv2 = read_inventory(buf)

        assert inv == inv2


@pytest.mark.usefixtures('ignore_numpy_errors')
@pytest.mark.skipif(not (CARTOPY_VERSION and CARTOPY_VERSION >= [0, 12, 0]),
                    reason='cartopy not installed')
class TestNetworkCartopy:
    """
    Tests for the :meth:`~obspy.station.network.Network.plot` with Cartopy.
    """
    @pytest.fixture(scope='class', autouse=True)
    def setup_teardown(self):
        """Perform class setup/teardown."""
        nperr = np.geterr()
        np.seterr(all='ignore')
        yield
        np.seterr(**nperr)

    def test_location_plot_global(self, image_path):
        """
        Tests the network location preview plot, default parameters, using
        Cartopy.
        """
        net = read_inventory()[0]
        net.plot(method='cartopy', outfile=image_path)

    def test_location_plot_ortho(self, image_path):
        """
        Tests the network location preview plot, ortho projection, some
        non-default parameters, using Cartopy.
        """
        net = read_inventory()[0]
        net.plot(method='cartopy', projection='ortho', resolution='c',
                 continent_fill_color='0.5', marker='d',
                 color='yellow', label=False, outfile=image_path)

    def test_location_plot_local(self, image_path):
        """
        Tests the network location preview plot, local projection, some more
        non-default parameters, using Cartopy.
        """
        net = read_inventory()[0]
        net.plot(method='cartopy', projection='local', resolution='50m',
                 size=13**2, outfile=image_path)



# -*- coding: utf-8 -*-
"""
The obspy.clients.earthworm.client test suite.
"""
import pytest

from obspy import read
from obspy.core.utcdatetime import UTCDateTime
from obspy.core.util import NamedTemporaryFile
from obspy.core.util.decorator import skip_on_network_error
from obspy.clients.earthworm import Client

pytestmark = pytest.mark.network


class TestEWClient:
    """
    Test cases for obspy.clients.earthworm.client.Client.
    """
    start = UTCDateTime() - 24 * 3600 * 4
    end = start + 1.0

    def try_get_stream(self, client, kwargs):
        """Try to get the stream with either '' or '--' location codes. """
        # I am not sure why, but sometimes the location code needs to be --
        # and other times '' in order to get a stream of nonzero length.
        # Just try both. Update: As of 11/2025 '' should work fine.
        st = client.get_waveforms(
            location='', starttime=self.start, endtime=self.end, **kwargs
        )
        if len(st) > 0:
            return st
        st = client.get_waveforms(
            location='--', starttime=self.start, endtime=self.end, **kwargs
        )
        return st

    @pytest.fixture(scope='class', autouse=True)
    def set_utc_precision(self):
        """Set UTC precision to 4 for this tests class."""
        current_precision = UTCDateTime.DEFAULT_PRECISION
        UTCDateTime.DEFAULT_PRECISION = 4
        yield
        UTCDateTime.DEFAULT_PRECISION = current_precision

    @pytest.fixture(scope='class')
    def ew_client(self):
        """Return the earthworn client."""
        ew_client = Client("pubavo1.wr.usgs.gov", 16022, timeout=30.0)
        return ew_client

    @pytest.fixture(scope='class')
    def ew_stream(self, ew_client):
        """Return a stream fetched from the test ew client."""

        # example 1 -- 1 channel, cleanup
        kwargs = dict(
            network='AV',
            station='AKV',
            channel='BHE',
        )
        return self.try_get_stream(ew_client, kwargs)

    @pytest.fixture(scope='class')
    def ew_stream_no_cleanup(self, ew_client):
        """Return a stream fetched from the test ew client with no cleanup."""
        kwargs = dict(
            network='AV',
            station='AKV',
            channel='BHE',
            cleanup=False,
        )
        return self.try_get_stream(ew_client, kwargs)

    @pytest.fixture(scope='class')
    def ew_stream_wildcard(self, ew_client):
        """Return a stream fetched from the test ew client with wildcard."""
        kwargs = dict(
            network='AV',
            station='AKV',
            channel='BH?',
        )
        return self.try_get_stream(ew_client, kwargs)

    @skip_on_network_error
    def test_get_waveform(self, ew_client, ew_stream):
        """
        Tests get_waveforms method.
        """
        assert len(ew_stream) == 1
        delta = ew_stream[0].stats.delta
        trace = ew_stream[0]
        assert len(trace) in (50, 51)
        assert trace.stats.starttime >= self.start - delta
        assert trace.stats.starttime <= self.start + delta
        assert trace.stats.endtime >= self.end - delta
        assert trace.stats.endtime <= self.end + delta
        assert trace.stats.network == 'AV'
        assert trace.stats.station == 'AKV'
        assert trace.stats.location in ('--', '')
        assert trace.stats.channel == 'BHE'

    @skip_on_network_error
    @pytest.mark.skip(reason='Example data AV.AKV..BHE no longer valid')
    def test_get_waveform_no_cleanup(self, ew_client, ew_stream_no_cleanup):
        """
        Tests get_waveforms method again, 1 channel no cleanup.
        """
        # example 2 -- 1 channel, no cleanup
        ew_stream = ew_stream_no_cleanup
        delta = ew_stream[0].stats.delta
        assert len(ew_stream) >= 2
        summed_length = sum(len(tr) for tr in ew_stream)
        assert summed_length in (50, 51)
        assert ew_stream[0].stats.starttime >= self.start - delta
        assert ew_stream[0].stats.starttime <= self.start + delta
        assert ew_stream[-1].stats.endtime >= self.end - delta
        assert ew_stream[-1].stats.endtime <= self.end + delta
        for trace in ew_stream:
            assert trace.stats.network == 'AV'
            assert trace.stats.station == 'AKV'
            assert trace.stats.location in ('--', '')
            assert trace.stats.channel == 'BHE'

    @skip_on_network_error
    def test_get_waveform_widlcard(self, ew_client, ew_stream_wildcard):
        """
        Test example 3 -- component wildcarded with '?'
        """
        # example 3 -- component wildcarded with '?'
        stream = ew_stream_wildcard
        delta = stream[0].stats.delta
        assert len(stream) == 3
        for trace in stream:
            assert len(trace) in (50, 51)
            assert trace.stats.starttime >= self.start - delta
            assert trace.stats.starttime <= self.start + delta
            assert trace.stats.endtime >= self.end - delta
            assert trace.stats.endtime <= self.end + delta
            assert trace.stats.network == 'AV'
            assert trace.stats.station == 'AKV'
            assert trace.stats.location in ('--', '')
        assert stream[0].stats.channel == 'BHZ'
        assert stream[1].stats.channel == 'BHN'
        assert stream[2].stats.channel == 'BHE'

    @skip_on_network_error
    def test_save_waveform(self, ew_client):
        """
        Tests save_waveforms method.
        """
        # initialize client
        with NamedTemporaryFile() as tf:
            testfile = tf.name
            # 1 channel, cleanup (using SLIST to avoid dependencies)
            ew_client.save_waveforms(
                testfile,
                'AV',
                'AKV',
                '--',
                'BHE',
                self.start,
                self.end,
                format="SLIST",
            )
            stream = read(testfile)
        assert len(stream) == 1
        delta = stream[0].stats.delta
        trace = stream[0]
        assert len(trace) == 51
        assert trace.stats.starttime >= self.start - delta
        assert trace.stats.starttime <= self.start + delta
        assert trace.stats.endtime >= self.end - delta
        assert trace.stats.endtime <= self.end + delta
        assert trace.stats.network == 'AV'
        assert trace.stats.station == 'AKV'
        assert trace.stats.location in ('--', '')
        assert trace.stats.channel == 'BHE'

    @skip_on_network_error
    def test_availability(self, ew_client):
        data = ew_client.get_availability()
        seeds = ["%s.%s.%s.%s" % (d[0], d[1], d[2], d[3]) for d in data]
        assert 'AV.AKV.--.BHZ' in seeds



# -*- coding: utf-8 -*-
"""
Module for map related plotting in ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import datetime
import warnings

import numpy as np
import matplotlib
from matplotlib.colorbar import Colorbar
from matplotlib.dates import AutoDateFormatter, AutoDateLocator, date2num
from matplotlib import patheffects
from matplotlib.ticker import (FormatStrFormatter, Formatter, FuncFormatter,
                               MaxNLocator)

from obspy import UTCDateTime
from obspy.core.util import CARTOPY_VERSION
from obspy.core.util.decorator import deprecated_keywords
from obspy.geodetics.base import mean_longitude

if CARTOPY_VERSION and CARTOPY_VERSION >= [0, 12, 0]:
    import cartopy
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature
    HAS_CARTOPY = True
else:
    HAS_CARTOPY = False

if not HAS_CARTOPY:
    msg = ("Cartopy not installed, map plots will not work.")
    warnings.warn(msg)


_CARTOPY_RESOLUTIONS = {
    'c': '110m',
    'l': '110m',
    'i': '50m',
    'h': '50m',
    'f': '10m',
    '110m': '110m',
    '50m': '50m',
    '10m': '10m',
}

if HAS_CARTOPY:
    _CARTOPY_FEATURES = {
        '110m': (cfeature.BORDERS, cfeature.LAND, cfeature.OCEAN),
    }


@deprecated_keywords({'bmap': None})
def _plot_cartopy_into_axes(
        ax, lons, lats, size, color, bmap=None, labels=None,
        projection='global', resolution='l', continent_fill_color='0.8',
        water_fill_color='1.0', colormap=None, marker="o", title=None,
        adjust_aspect_to_colorbar=False, **kwargs):  # @UnusedVariable

    """
    Creates a (or adds to existing) cartopy plot with a data point scatter
    plot in given axes.

    See :func:`plot_cartopy` for details on most args/kwargs.


    :type ax: :class:`matplotlib.axes.Axes` or
        :class:`cartopy.mpl.geoaxes.GeoAxes`
    :param ax: Existing matplotlib axes instance, optionally with previous
        cartopy plot. If a cartopy GeoAxes is provided, most setup steps will
        be skipped.
    :type bmap: :class:`matplotlib.axes.Axes`
    :param bmap: Deprecated and unused. Whether `ax` is a plain matplotlib Axes
        or a cartopy GeoAxes will determine if cartopy related setup on the
        axis is skipped (setting up projection etc.).
    :rtype: :class:`matplotlib.collections.PathCollection`
    :returns: Matplotlib path collection (e.g. to reuse for colorbars).
    """

    if not isinstance(ax, cartopy.mpl.geoaxes.GeoAxes):
        if projection in ['global', 'ortho']:
            pass
        elif projection == 'local':
            if min(lons) < -150 and max(lons) > 150:
                max_lons = max(np.array(lons) % 360)
                min_lons = min(np.array(lons) % 360)
            else:
                max_lons = max(lons)
                min_lons = min(lons)

            ax.set_extent([min_lons, max_lons, min(lats), max(lats)])

        else:
            msg = "Projection '%s' not supported." % projection
            raise ValueError(msg)
        # ax.gridlines()
        # ax.coastlines()
        # draw coast lines, country boundaries, fill continents.
        # ax.set_facecolor(water_fill_color)
        # newer matplotlib errors out if called with empty coastline data (no
        # coast on map)
        # if np.size(getattr(bmap, 'coastsegs', [])):
        #     bmap.drawcoastlines(color="0.4")
        # bmap.drawcountries(color="0.75")
        # bmap.fillcontinents(color=continent_fill_color,
        #                     lake_color=water_fill_color)
        # draw the edge of the bmap projection region (the projection limb)
        # bmap.drawmapboundary(fill_color=water_fill_color)
        # draw lat/lon grid lines every 30 degrees.
        # bmap.drawmeridians(np.arange(-180, 180, 30))
        # bmap.drawparallels(np.arange(-90, 90, 30))
        ax.stock_img()
        ax.gridlines()
        ax.coastlines()

    # compute the native bmap projection coordinates for events.
    # x, y = bmap(lons, lats)
    x, y = (lons, lats)
    # plot labels

    if labels:
        if 100 > len(lons) > 1:
            for name, xpt, ypt, _colorpt in zip(labels, x, y, color):
                # Check if the point can actually be seen with the current bmap
                # projection. The bmap object will set the coordinates to very
                # large values if it cannot project a point.
                if xpt > 1e25:
                    continue
                ax.text(xpt, ypt, name, weight="heavy",
                        color="k", zorder=100,
                        path_effects=[
                            patheffects.withStroke(linewidth=3,
                                                   foreground="white")],
                        transform=ccrs.Geodetic())
        elif len(lons) == 1:
            ax.text(x[0], y[0], labels[0], weight="heavy", color="k",
                    path_effects=[
                        patheffects.withStroke(linewidth=3,
                                               foreground="white")],
                    transform=ccrs.Geodetic())

    # scatter plot is removing valid x/y points with invalid color value,
    # so we plot those points separately.
    try:
        nan_points = np.isnan(np.array(color, dtype=float))
    except ValueError:
        # `color' was not a list of values, but a list of colors.
        pass
    else:
        if nan_points.any():
            x_ = np.array(x)[nan_points]
            y_ = np.array(y)[nan_points]
            size_ = np.array(size)[nan_points]
            ax.scatter(x_, y_, marker=marker, s=size_, c="0.3",
                       zorder=10, cmap=None, transform=ccrs.Geodetic())
    # Had to change transform to ccrs.PlateCarree, see:
    # https://stackoverflow.com/a/13657749/3645626
    scatter = ax.scatter(x, y, marker=marker, s=size, c=color, zorder=10,
                         cmap=colormap, transform=ccrs.PlateCarree(),)

    if title:
        ax.set_title(title)

    return scatter


def plot_cartopy(lons, lats, size, color, labels=None, projection='global',
                 resolution='110m', continent_fill_color='0.8',
                 water_fill_color='1.0', colormap=None, colorbar=None,
                 marker="o", title=None, colorbar_ticklabel_format=None,
                 show=True, proj_kwargs=None, ax=None,
                 **kwargs):  # @UnusedVariable
    """
    Creates a Cartopy plot with a data point scatter plot.

    :type lons: list[float] or tuple(float)
    :param lons: Longitudes of the data points.
    :type lats: list[float] or tuple(float)
    :param lats: Latitudes of the data points.
    :type size: float, list[float] or tuple(float)
    :param size: Size of the individual points in the scatter plot.
    :type color: list[float], tuple(float) or objects that can be
        converted to floats, like e.g.
        :class:`~obspy.core.utcdatetime.UTCDateTime`)
    :param color: Color information of the individual data points to be
        used in the specified color map (e.g. origin depths,
        origin times).
    :type labels: list[str] or tuple[float]
    :param labels: Annotations for the individual data points.
    :type projection: str, optional
    :param projection: The map projection.
        Currently supported are:

            * ``"global"`` (Will plot the whole world using
              :class:`~cartopy.crs.Mollweide`.)
            * ``"ortho"`` (Will center around the mean lat/long using
              :class:`~cartopy.crs.Orthographic`.)
            * ``"local"`` (Will plot around local events using
              :class:`~cartopy.crs.AlbersEqualArea`.)
            * Any other Cartopy :class:`~cartopy.crs.Projection`. An instance
              of this class will be created using the supplied ``proj_kwargs``.

        Defaults to "global"
    :type resolution: str, optional
    :param resolution: Resolution of the boundary database to use. Will be
        passed directly to the Cartopy module. Possible values are:

            * ``"110m"``
            * ``"50m"``
            * ``"10m"``

        Defaults to ``"110m"``. For compatibility, you may also specify any of
        the cartopy resolutions defined in :func:`plot_cartopy`.
    :type continent_fill_color: valid matplotlib color, optional
    :param continent_fill_color:  Color of the continents. Defaults to
        ``"0.9"`` which is a light gray.
    :type water_fill_color: valid matplotlib color, optional
    :param water_fill_color: Color of all water bodies.
        Defaults to ``"white"``.
    :type colormap: str, valid matplotlib colormap, optional
    :param colormap: The colormap for color-coding the events as provided
        in `color` kwarg.
        The event with the smallest `color` property will have the
        color of one end of the colormap and the event with the highest
        `color` property the color of the other end with all other events
        in between.
        Defaults to None which will use the default matplotlib colormap.
    :type colorbar: bool, optional
    :param colorbar: When left `None`, a colorbar is plotted if more than one
        object is plotted. Using `True`/`False` the colorbar can be forced
        on/off.
    :type title: str
    :param title: Title above plot.
    :type colorbar_ticklabel_format: str or function or
        subclass of :class:`matplotlib.ticker.Formatter`
    :param colorbar_ticklabel_format: Format string or Formatter used to format
        colorbar tick labels.
    :type show: bool
    :param show: Whether to show the figure after plotting or not. Can be used
        to do further customization of the plot before showing it.
    :type proj_kwargs: dict
    :param proj_kwargs: Keyword arguments to pass to the Cartopy
        :class:`~cartopy.crs.Projection`. In this dictionary, you may specify
        ``central_longitude='auto'`` or ``central_latitude='auto'`` to have
        this function calculate the latitude or longitude as it would for other
        projections. Some arguments may be ignored if you choose one of the
        built-in ``projection`` choices.
    :type ax: :class:`matplotlib.axes.Axes` or
        :class:`cartopy.mpl.geoaxes.GeoAxes`
    :param ax: Existing matplotlib axes instance, optionally with previous
        cartopy plot. If a cartopy GeoAxes is provided, most setup steps will
        be skipped.
    """
    import matplotlib.pyplot as plt

    if isinstance(color[0], (datetime.datetime, UTCDateTime)):
        datetimeplot = True
        color = [date2num(getattr(t, 'datetime', t)) for t in color]
    else:
        datetimeplot = False

    # If ax wasn't specified, look for fig in kwargs
    if ax is None and kwargs.get("fig"):
        ax = kwargs['fig'].axes[0]

    if ax is None:
        fig, map_ax, cm_ax, show_colorbar = _basic_setup(
            lons=lons, lats=lats, size=size, color=color, labels=labels,
            projection=projection, resolution=resolution,
            continent_fill_color=continent_fill_color,
            water_fill_color=water_fill_color, colormap=colormap,
            colorbar=colorbar, marker=marker, title=title,
            colorbar_ticklabel_format=colorbar_ticklabel_format,
            proj_kwargs=proj_kwargs)
    else:
        if isinstance(ax, matplotlib.figure.Figure):
            fig = ax
            map_ax = fig.axes[0]
        else:
            fig = ax.figure
            map_ax = ax
        cm_ax = None
        show_colorbar = False

    # Plot labels
    if labels and len(lons) > 0:
        with map_ax.hold_limits():
            for name, xpt, ypt, _colorpt in zip(labels, lons, lats, color):
                map_ax.text(xpt, ypt, name, weight="heavy", color="k",
                            zorder=100, transform=ccrs.PlateCarree(),
                            path_effects=[
                                patheffects.withStroke(linewidth=3,
                                                       foreground="white")])

    scatter = map_ax.scatter(lons, lats, marker=marker, s=size, c=color,
                             zorder=10, cmap=colormap,
                             transform=ccrs.PlateCarree())

    if title:
        fig.suptitle(title)

    # Only show the colorbar for more than one event.
    if show_colorbar:
        if colorbar_ticklabel_format is not None:
            if isinstance(colorbar_ticklabel_format, str):
                formatter = FormatStrFormatter(colorbar_ticklabel_format)
            elif hasattr(colorbar_ticklabel_format, '__call__'):
                formatter = FuncFormatter(colorbar_ticklabel_format)
            elif isinstance(colorbar_ticklabel_format, Formatter):
                formatter = colorbar_ticklabel_format
            locator = MaxNLocator(5)
        else:
            if datetimeplot:
                locator = AutoDateLocator()
                formatter = AutoDateFormatter(locator)
                # Compat with old matplotlib versions.
                if hasattr(formatter, "scaled"):
                    formatter.scaled[1 / (24. * 60.)] = '%H:%M:%S'
            else:
                locator = None
                formatter = None
        cb = Colorbar(cm_ax, scatter,
                      orientation='horizontal',
                      ticks=locator,
                      format=formatter)
        # Compat with old matplotlib versions.
        if hasattr(cb, "update_ticks"):
            cb.update_ticks()

    if show:
        plt.show()

    return fig


def _basic_setup(
        lons, lats, size, color, labels, projection, resolution,
        continent_fill_color, water_fill_color, colormap, colorbar, marker,
        title, colorbar_ticklabel_format, proj_kwargs):
    import matplotlib.pyplot as plt

    fig = plt.figure()

    # The colorbar should only be plotted if more then one event is
    # present.
    if colorbar is not None:
        show_colorbar = colorbar
    else:
        if len(lons) > 1 and hasattr(color, "__len__") and \
                not isinstance(color, str):
            show_colorbar = True
        else:
            show_colorbar = False

    if projection == "local":
        ax_x0, ax_width = 0.10, 0.80
    elif projection == "global":
        ax_x0, ax_width = 0.01, 0.98
    else:
        ax_x0, ax_width = 0.05, 0.90

    proj_kwargs = proj_kwargs or {}
    if projection == 'global':
        proj_kwargs['central_longitude'] = np.mean(lons)
        proj = ccrs.Mollweide(**proj_kwargs)
    elif projection == 'ortho':
        proj_kwargs['central_latitude'] = np.mean(lats)
        proj_kwargs['central_longitude'] = mean_longitude(lons)
        proj = ccrs.Orthographic(**proj_kwargs)
    elif projection == 'local':
        if min(lons) < -150 and max(lons) > 150:
            max_lons = max(np.array(lons) % 360)
            min_lons = min(np.array(lons) % 360)
        else:
            max_lons = max(lons)
            min_lons = min(lons)
        lat_0 = max(lats) / 2. + min(lats) / 2.
        lon_0 = max_lons / 2. + min_lons / 2.
        if lon_0 > 180:
            lon_0 -= 360
        deg2m_lat = 2 * np.pi * 6371 * 1000 / 360
        deg2m_lon = deg2m_lat * np.cos(lat_0 / 180 * np.pi)
        if len(lats) > 1:
            height = (max(lats) - min(lats)) * deg2m_lat
            width = (max_lons - min_lons) * deg2m_lon
            margin = 0.2 * (width + height)
            height += margin
            width += margin
        else:
            height = 2.0 * deg2m_lat
            width = 5.0 * deg2m_lon
        # Do intelligent aspect calculation for local projection
        # adjust to figure dimensions
        w, h = fig.get_size_inches()
        aspect = w / h
        if show_colorbar:
            aspect *= 1.2
        if width / height < aspect:
            width = height * aspect
        else:
            height = width / aspect

        proj_kwargs['central_latitude'] = lat_0
        proj_kwargs['central_longitude'] = lon_0
        proj_kwargs['standard_parallels'] = [lat_0, lat_0]
        proj = ccrs.AlbersEqualArea(**proj_kwargs)

    # User-supplied projection.
    elif isinstance(projection, type):
        if 'central_longitude' in proj_kwargs:
            if proj_kwargs['central_longitude'] == 'auto':
                proj_kwargs['central_longitude'] = mean_longitude(lons)
        if 'central_latitude' in proj_kwargs:
            if proj_kwargs['central_latitude'] == 'auto':
                proj_kwargs['central_latitude'] = np.mean(lats)
        if 'pole_longitude' in proj_kwargs:
            if proj_kwargs['pole_longitude'] == 'auto':
                proj_kwargs['pole_longitude'] = np.mean(lons)
        if 'pole_latitude' in proj_kwargs:
            if proj_kwargs['pole_latitude'] == 'auto':
                proj_kwargs['pole_latitude'] = np.mean(lats)

        proj = projection(**proj_kwargs)

    else:
        msg = "Projection '%s' not supported." % projection
        raise ValueError(msg)

    if show_colorbar:
        map_ax = fig.add_axes([ax_x0, 0.13, ax_width, 0.77], projection=proj)
        cm_ax = fig.add_axes([ax_x0, 0.05, ax_width, 0.05])
        plt.sca(map_ax)
    else:
        ax_y0, ax_height = 0.05, 0.85
        if projection == "local":
            ax_y0 += 0.05
            ax_height -= 0.05
        map_ax = fig.add_axes([ax_x0, ax_y0, ax_width, ax_height],
                              projection=proj)
        cm_ax = None

    if projection == 'local':
        x0, y0 = proj.transform_point(lon_0, lat_0, proj.as_geodetic())
        map_ax.set_xlim(x0 - width / 2, x0 + width / 2)
        map_ax.set_ylim(y0 - height / 2, y0 + height / 2)
    else:
        map_ax.set_global()

    # Pick features at specified resolution.
    resolution = _CARTOPY_RESOLUTIONS[resolution]
    try:
        borders, land, ocean = _CARTOPY_FEATURES[resolution]
    except KeyError:
        borders = cfeature.NaturalEarthFeature(
            cfeature.BORDERS.category, cfeature.BORDERS.name, resolution,
            edgecolor='none', facecolor='none')
        land = cfeature.NaturalEarthFeature(
            cfeature.LAND.category, cfeature.LAND.name, resolution,
            edgecolor='face', facecolor='none')
        ocean = cfeature.NaturalEarthFeature(
            cfeature.OCEAN.category, cfeature.OCEAN.name, resolution,
            edgecolor='face', facecolor='none')
        _CARTOPY_FEATURES[resolution] = (borders, land, ocean)

    # Draw coast lines, country boundaries, fill continents.
    map_ax.set_facecolor(water_fill_color)
    map_ax.add_feature(ocean, facecolor=water_fill_color)
    map_ax.add_feature(land, facecolor=continent_fill_color)
    map_ax.add_feature(borders, edgecolor='0.75')
    map_ax.coastlines(resolution=resolution, color='0.4')

    # Draw grid lines - TODO: draw_labels=True doesn't work yet.
    if projection == 'local':
        map_ax.gridlines()
    else:
        # Draw lat/lon grid lines every 30 degrees.
        map_ax.gridlines(xlocs=range(-180, 181, 30), ylocs=range(-90, 91, 30))

    return fig, map_ax, cm_ax, show_colorbar


def plot_map(method, *args, **kwargs):
    """
    Creates a map plot with a data point scatter plot.

    :type method: str
    :param method: Method to use for plotting. Possible values are:

        * ``'cartopy'`` to use the Cartopy library. For other arguments, see
          the :func:`plot_cartopy` function.
        * ``None`` will use the Cartopy library since it is the only supported
          method right now.
    """
    if method is None:
        if HAS_CARTOPY:
            return plot_cartopy(*args, **kwargs)
        else:
            raise ImportError('Cartopy could not be imported.')
    elif method == 'cartopy':
        if not HAS_CARTOPY:
            raise ImportError('Cartopy cannot be imported but was explicitly '
                              'requested.')
        return plot_cartopy(*args, **kwargs)
    else:
        raise ValueError("The method argument must be either 'None' or "
                         "'cartopy', not '%s'." % (method, ))



import numpy as np
import matplotlib.pyplot as plt

from obspy.imaging.cm import obspy_sequential
from obspy.signal.array_analysis import array_transff_wavenumber


# generate array coordinates
coords = np.array([[10., 60., 0.], [200., 50., 0.], [-120., 170., 0.],
                   [-100., -150., 0.], [30., -220., 0.]])

# coordinates in km
coords /= 1000.

# set limits for wavenumber differences to analyze
klim = 40.
kxmin = -klim
kxmax = klim
kymin = -klim
kymax = klim
kstep = klim / 100.

# compute transfer function as a function of wavenumber difference
transff = array_transff_wavenumber(coords, klim, kstep, coordsys='xy')

# plot
plt.pcolor(np.arange(kxmin, kxmax + kstep * 1.1, kstep) - kstep / 2.,
           np.arange(kymin, kymax + kstep * 1.1, kstep) - kstep / 2.,
           transff.T, cmap=obspy_sequential)

plt.colorbar()
plt.clim(vmin=0., vmax=1.)
plt.xlim(kxmin, kxmax)
plt.ylim(kymin, kymax)
plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
seiscomp.core test suite.

:author:
    EOST (École et Observatoire des Sciences de la Terre)
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import re

import pytest

from obspy.io.seiscomp.core import _is_scml, validate


class TestCore():
    """
    Test suite for obspy.io.seiscomp.event
    """
    def test_scml_versions(self, testdata):
        """
        Test multiple schema versions
        """
        for version in ['0.10', '0.11', '0.12', '0.13', '0.14']:
            filename = testdata['version%s' % version]
            assert _is_scml(filename)

    def test_scml_no_version_attribute(self, testdata):
        filename = testdata['no_version_attribute.sc3ml']
        assert _is_scml(filename)

    def test_validate(self, testdata):
        filename = testdata['qml-example-1.2-RC3.sc3ml']
        assert validate(filename)
        assert not validate(filename, version='0.8')

        expected_error = re.escape(
            "0.99 is not a supported version. Use one of these"
            " versions: [0.7, 0.8, 0.9, 0.10, "
            "0.11, 0.12, 0.13, 0.14].")
        with pytest.raises(ValueError, match=expected_error):
            validate(filename, version='0.99')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests the SeismicPhase class.
"""
import numpy as np

from obspy.taup.tau_model import TauModel


class TestSplitTauModel:
    """
    Test suite for splitting of the TauModel class.
    """

    def test_split(self):
        depth = 110
        tau_model = TauModel.from_file('iasp91')
        split_t_mod = tau_model.split_branch(depth)
        shape1 = tau_model.tau_branches.shape[1] + 1
        shape2 = split_t_mod.tau_branches.shape[1]
        assert shape1 == shape2

        assert len(tau_model.ray_params) + 2 == len(split_t_mod.ray_params)

        branch_count = tau_model.tau_branches.shape[1]
        split_branch_index = tau_model.find_branch(depth)

        new_p_ray_param = split_t_mod.s_mod.get_slowness_layer(
            split_t_mod.s_mod.layer_number_above(depth, True), True)['bot_p']
        new_s_ray_param = split_t_mod.s_mod.get_slowness_layer(
            split_t_mod.s_mod.layer_number_above(depth, False), False)['bot_p']

        p_index = s_index = len(split_t_mod.ray_params)
        ray_params = split_t_mod.ray_params
        for j in range(len(ray_params)):
            if new_p_ray_param == ray_params[j]:
                p_index = j
            if new_s_ray_param == ray_params[j]:
                s_index = j

        assert p_index == len(split_t_mod.ray_params) or s_index < p_index

        for b in range(branch_count):
            orig = tau_model.get_tau_branch(b, True)
            if b < split_branch_index:
                depth_branch = split_t_mod.get_tau_branch(b, True)
                assert depth_branch.dist[p_index] > 0
                assert depth_branch.time[p_index] > 0
            elif b > split_branch_index:
                depth_branch = split_t_mod.get_tau_branch(b + 1, True)
                assert np.isclose(depth_branch.dist[p_index], 0)
                assert np.isclose(depth_branch.time[p_index], 0)
            else:
                # the split one
                continue

            np.testing.assert_allclose(orig.dist[0:s_index],
                                       depth_branch.dist[0:s_index],
                                       atol=0.00000001)
            np.testing.assert_allclose(orig.time[0:s_index],
                                       depth_branch.time[0:s_index],
                                       atol=0.00000001)
            orig_len = len(orig.dist)
            if s_index < orig_len:
                assert orig_len + 2 == len(depth_branch.dist)
                np.testing.assert_allclose(
                    orig.dist[s_index:p_index - 1],
                    depth_branch.dist[s_index + 1:p_index],
                    atol=0.00000001)
                np.testing.assert_allclose(
                    orig.time[s_index:p_index - 1],
                    depth_branch.time[s_index + 1:p_index],
                    atol=0.00000001)
                np.testing.assert_allclose(
                    orig.dist[p_index:orig_len - s_index - 2],
                    depth_branch.dist[p_index + 2:orig_len - s_index],
                    atol=0.00000001)
                np.testing.assert_allclose(
                    orig.time[p_index:orig_len - s_index - 2],
                    depth_branch.time[p_index + 2:orig_len - s_index],
                    atol=0.00000001)

        # now check branch split
        orig = tau_model.get_tau_branch(split_branch_index, True)
        above = split_t_mod.get_tau_branch(split_branch_index, True)
        below = split_t_mod.get_tau_branch(split_branch_index + 1, True)
        assert np.isclose(above.min_ray_param, below.max_ray_param)
        for i in range(len(above.dist)):
            if i < s_index:
                assert np.isclose(orig.dist[i], above.dist[i])
            elif i == s_index:
                # new value should be close to average of values to either side
                assert np.isclose((orig.dist[i - 1] + orig.dist[i]) / 2,
                                  above.dist[i], atol=0.00001)
            elif s_index < i < p_index:
                assert np.isclose(orig.dist[i - 1],
                                  above.dist[i] + below.dist[i])
            elif i == p_index:
                # new value should be close to average of values to either side
                val1 = (orig.dist[i - 2] + orig.dist[i - 1]) / 2
                val2 = above.dist[i] + below.dist[i]
                assert np.isclose(val1, val2, atol=0.0001)
            else:
                val1 = orig.dist[i - 2]
                val2 = above.dist[i] + below.dist[i]
                assert np.isclose(val1, val2, atol=0.000000001)
